#!/usr/bin/env python3

import os, re, shutil, logging
import numpy as np
import subprocess as sp
import hashlib, pathlib

iter_format = "%06d"
task_format = "%02d"
log_iter_head = "iter " + iter_format + " task " + task_format + ": "
float_protect = 1e-14

def make_iter_name (iter_index) :
    return "iter." + (iter_format % iter_index)


def get_first_matched_key_from_dict(dct, lst):
    value = None
    for key in lst:
        if key in dct:
            value = dct[key]
            return value
    raise KeyError(f"not found key in dct={dct}, lst={lst}")


def create_dict_not_empty_key(**kwargs):
    dct = {}
    for k,v in kwargs.items():
        if v is not None:
            dct[k] = v
    return dct

# def create_path (path) :
#     path += '/'
#     if os.path.isdir(path) : 
#         dirname = os.path.dirname(path)        
#         counter = 0
#         while True :
#             bk_dirname = dirname + ".bk%03d" % counter
#             if not os.path.isdir(bk_dirname) : 
#                 shutil.move (dirname, bk_dirname) 
#                 break
#             counter += 1
#     os.makedirs (path)





def create_path (path) :
    path += '/'
    if os.path.isdir(path) :
        dirname = os.path.dirname(path)
        counter = 0
        while True :
            bk_dirname = dirname + ".bk%03d" % counter
            if not os.path.isdir(bk_dirname) :
                shutil.move (dirname, bk_dirname)
                break
            counter += 1
    os.makedirs (path)
    abs_path = os.path.abspath(path)
    return abs_path


def relative_link_file(file_path, target_dir):
    if not os.path.isfile(file_path):
        raise RuntimeError(f"file_path:{file_path} must be a file. cwd:{os.getcwd()}")
    file_abs_path = os.path.abspath(file_path)
    basename = os.path.basename(file_abs_path)
    relative_path = os.path.relpath(file_abs_path, start=target_dir)
    target_linkfile_path = os.path.join(target_dir, basename)
    os.symlink(src=relative_path, dst=target_linkfile_path)
    return target_linkfile_path

def link_file_in_dict(dct, 
        key_list, 
        target_dir
    ):
    if not dct:
        return {}
    return_dict = {}
    for k in key_list:
        file_path = dct.get(k, None)
        if file_path is not None:
            target_linkfile_path = relative_link_file(
                file_path=file_path,
                target_dir=target_dir
            )
            v = os.path.basename(target_linkfile_path)
            return_dict[k] = v
    return return_dict

def get_file_md5(file_path):
    return hashlib.md5(pathlib.Path(file_path).read_bytes()).hexdigest()

def copy_file_list (file_list, from_path, to_path) :
    for jj in file_list : 
        if os.path.isfile(os.path.join(from_path, jj)) :
            shutil.copy (os.path.join(from_path, jj), to_path)
        elif os.path.isdir(os.path.join(from_path, jj)) :
            shutil.copytree (os.path.join(from_path, jj), os.path.join(to_path, jj))


def block_avg(inp, skip = 0, block_size = 10) :
    inp = inp[skip:]
    data_chunks = [
        [j for j in inp[i:i+block_size]] \
        for i in range(0, len(inp), block_size)
    ]
    nblocks = len(data_chunks)
    if (len(data_chunks[-1]) != block_size) :
        nblocks -= 1
        data_chunks = data_chunks[:nblocks]
    assert (len(data_chunks) == nblocks)
    # naive avg
    naive_avg = np.average(inp)
    naive_err = np.std(inp) / np.sqrt(len(inp) - 1)
    # block avg
    data_chunks = np.array(data_chunks)
    data_block = np.average(data_chunks, axis = 1)
    block_avg = np.average(data_block)
    if len(data_block) != 1 :
        block_err = np.std(data_block) / np.sqrt(nblocks-1)
    else :
        block_err = None

    return block_avg, block_err

def cvt_conf (fin, 
              fout, 
              ofmt = 'vasp') :
    """
    Format convert from fin to fout, specify the output format by ofmt
    """
    thisfile = os.path.abspath(__file__)
    thisdir = os.path.dirname(thisfile)
    cmd = os.path.join(thisdir, 'ovito_file_convert.py')
    cmd_opt = '-m '+ofmt
    cmd_line = cmd + ' ' + cmd_opt + ' ' + fin + ' ' + fout
    sp.check_call(cmd_line, shell = True)    
    # sp.check_call([cmd, cmd_opt, fin, fout])

def _parse_one_str(in_s) :
    fmt_s = in_s.split(':') 
    if len(fmt_s) == 1 :
        return np.array([float(fmt_s[0])])
    else :
        assert(len(fmt_s)) == 3 
        # print(float(fmt_s[0]), float(fmt_s[1]), float(fmt_s[2]))
        return np.arange(float(fmt_s[0]),
                         float(fmt_s[1]) - float_protect, 
                         float(fmt_s[2]))

def parse_seq(in_s, *, protect_eps=None):
    all_l = []
    if type(in_s) == list and type(in_s[0]) == str :
        for ii in in_s :
            for jj in _parse_one_str(ii) :
                all_l.append(jj)      
    elif type(in_s) == list and \
         ( type(in_s[0]) == float or \
           type(in_s[0]) == np.float32 or \
           type(in_s[0]) == np.float64 or \
           type(in_s[0]) == int ) :
        all_l = [float(ii) for ii in in_s]
    elif type(in_s) == str :
        all_l = _parse_one_str(in_s)
    else :
        raise RuntimeError("the type of seq should be one of: string, list_of_strings, list_of_floats")
    if protect_eps is not None:
        if all_l[0] == 0 :
            all_l[0] += protect_eps
        if all_l[-1] == 1 :
            all_l[-1] -= protect_eps
    # return all_l
    return np.array(all_l)

# def integrate(xx, yy) :
#     diff_e = 0
#     ntasks = len(xx) - 1
#     assert (len(yy) - 1 == ntasks)
#     for ii in range(ntasks) :
#         diff_e += 0.5 * (xx[ii+1] - xx[ii]) * (yy[ii+1] + yy[ii])
#     return (diff_e)
    
def integrate_trapezoidal(xx, yy, ye) :
    diff_e = 0
    err = 0
    ntasks = len(xx) - 1
    assert (len(yy) - 1 == ntasks)
    for ii in range(ntasks) :
        diff_e += 0.5 * (xx[ii+1] - xx[ii]) * (yy[ii+1] + yy[ii])
        err += np.square(0.5 * (xx[ii+1] - xx[ii]) * ye[ii+1])
        err += np.square(0.5 * (xx[ii+1] - xx[ii]) * ye[ii])
    return diff_e, np.sqrt(err)

def integrate_simpson(xx, yy, ye):
    if len(xx) % 2 == 0:
        diff_e, err = integrate_simpson(xx[:-1], yy[:-1], ye[:-1])
        diff_e1, err1 = integrate_trapezoidal(xx[-2:], yy[-2:], ye[-2:])
        return diff_e+diff_e1, np.linalg.norm([err, err1])
    else:
        diff_e = 0
        err = 0
        for ii in range(0, len(xx)-2, 2):
            two_h = (xx[ii+2] - xx[ii])
            diff_e += 1./6. * two_h * (yy[ii] + 4.*yy[ii+1] + yy[ii+2])
            err += np.square(1./6. * two_h * 1. * ye[ii])
            err += np.square(1./6. * two_h * 4. * ye[ii+1])
            err += np.square(1./6. * two_h * 1. * ye[ii+2])
        return diff_e, np.sqrt(err)

def integrate_simpson_nonuniform(x, f, fe):
    N = len(x) - 1
    h = np.diff(x)
    result = 0.0
    error = 0.0
    for i in range(1, N, 2):
        hph = h[i] + h[i-1]
        pref1 = ( h[i]**3 + h[i-1]**3 + 3. * h[i] * h[i-1] * hph ) / ( 6 * h[i] * h[i-1] )
        pref0 = ( 2. * h[i-1]**3 - h[i]**3 + 3. * h[i] * h[i-1]**2) / ( 6 * h[i-1] * hph)
        pref2 = ( 2. * h[i]**3 - h[i-1]**3 + 3. * h[i-1] * h[i]**2) / ( 6 * h[i] * hph )
        result += f[i  ] * pref1
        result += f[i-1] * pref0
        result += f[i+1] * pref2
        error += np.square(fe[i  ] * pref1)
        error += np.square(fe[i-1] * pref0)
        error += np.square(fe[i+1] * pref2)
    if (N + 1) % 2 == 0:
        pref1 = ( 2 * h[N-1]**2 + 3. * h[N-2] * h[N-1]) / ( 6 * ( h[N-2] + h[N-1] ) )
        pref0 = ( h[N-1]**2 + 3*h[N-1]* h[N-2] ) / ( 6 * h[N-2] )
        pref2 = h[N-1]**3 / ( 6 * h[N-2] * ( h[N-2] + h[N-1] ) )
        result += f[N  ] * pref1
        result += f[N-1] * pref0
        result -= f[N-2] * pref2
        error += np.square(fe[N  ] * pref1)
        error += np.square(fe[N-1] * pref0)
        error += np.square(fe[N-2] * pref2)
    return result, np.sqrt(error)

def integrate(xx, yy, ye, scheme_ = 's'):
    scheme = (scheme_.lower()[0])
    if scheme == 't':
        return integrate_trapezoidal(xx, yy, ye)
    elif scheme == 's':
        return integrate_simpson_nonuniform(xx, yy, ye)
    else:
        raise RuntimeError('unknow integration scheme', scheme_)

def _interval_deriv2 (xx, yy) :
    mat = np.ones([3, 3])
    for ii in range(3) :
        mat[ii][0] = xx[ii]**2
        mat[ii][1] = xx[ii]
    coeff = np.linalg.solve(mat, yy)
    return 2.*coeff[0]

def interval_sys_err_trapezoidal (xx, yy, mode) :
    if mode == 'middle' :        
        d0 = np.abs(_interval_deriv2(xx[0:3], yy[0:3]))
        d1 = np.abs(_interval_deriv2(xx[1:4], yy[1:4]))
        dd = np.max([d0, d1])
        dx = np.abs(xx[2] - xx[1])
    elif mode == 'left' :
        dd = np.abs(_interval_deriv2(xx[0:3], yy[0:3]))
        dx = np.abs(xx[1] - xx[0])
    elif mode == 'right' :
        dd = np.abs(_interval_deriv2(xx[0:3], yy[0:3]))
        dx = np.abs(xx[2] - xx[1])
    return 1./12.*(dx**3)*dd


def integrate_sys_err_trapezoidal (xx, yy) :
    err = 0
    if len(xx) <= 2 :
        return err
    err += interval_sys_err_trapezoidal(xx[0:3], yy[0:3], 'left')
    # print('here', err)
    for ii in range(1, len(xx)-2) :
        err += interval_sys_err_trapezoidal(xx[ii-1:ii+3], yy[ii-1:ii+3], 'middle')
        # print('here2', interval_sys_err(xx[ii-1:ii+3], yy[ii-1:ii+3], 'middle'))
    err += interval_sys_err_trapezoidal(xx[-3:], yy[-3:], 'right')    
    # print('here1', interval_sys_err(xx[-3:], yy[-3:], 'right') )
    return err


def integrate_sys_err_simpson (xx, yy) :
    err = 0
    if len(xx) <= 4 :
        return err
    xx = np.array(xx)
    ye = np.zeros(xx.size)
    interval_error = np.zeros(xx.size//4+1)
    for ii in range(xx.size//4):
        inte0,_ = integrate_simpson_nonuniform(xx[ii*4:ii*4+5], yy[ii*4:ii*4+5], ye[ii*4:ii*4+5])
        inte1,_ = integrate_simpson_nonuniform(xx[ii*4:ii*4+5:2], yy[ii*4:ii*4+5:2], ye[ii*4:ii*4+5:2])
        err = np.abs(inte0 - inte1)
        interval_error[ii+1] = err
    # from scipy.interpolate import interp1d
    # err = interp1d(xx[::4], err0, kind='cubic')
    return interval_error[-1]


def integrate_sys_err (xx, yy, scheme_ = 's') :
    scheme = (scheme_.lower()[0])
    if scheme == 't':
        return integrate_sys_err_trapezoidal(xx, yy)
    elif scheme == 's':
        return integrate_sys_err_simpson(xx, yy)
    else:
        raise RuntimeError('unknow integration scheme', scheme_)

def integrate_range_trapezoidal(xx, yy, ye):
    xx = np.array(xx)
    nn = xx.size
    inte = np.zeros([nn])
    stat_err = np.zeros([nn])
    inte_err = np.zeros([nn])
    # stat_err[0] = ye[0]
    for ii in range(1, nn):
        inter_i, inter_se = integrate_trapezoidal(xx[ii-1:ii+1], yy[ii-1:ii+1], ye[ii-1:ii+1])
        inte[ii] = inte[ii-1] + inter_i
        stat_err[ii] = np.sqrt(stat_err[ii-1]**2 + inter_se**2)
    if nn <= 2:
        return xx, inte, inte_err, stat_err
    inte_err[1] = interval_sys_err_trapezoidal(xx[0:3], yy[0:3], 'left')
    for ii in range(1, nn-2):
        inte_err[ii+1] = inte_err[ii] + interval_sys_err_trapezoidal(xx[ii-1:ii+3], yy[ii-1:ii+3], 'middle')
    inte_err[nn-1] = inte_err[nn-2] + interval_sys_err_trapezoidal(xx[-3:], yy[-3:], 'right')
    return xx, inte, inte_err, stat_err

def _integrate_range_simpson_inner(xx, yy, ye):
    xx = np.array(xx)
    nn = xx.size
    new_xx = [xx[0]]
    inte = [0]
    # stat_err = [ye[0]]
    stat_err = [0]
    inte_err = [0]
    for ii in range(2, nn, 2):
        inter_i, inter_se = integrate_simpson_nonuniform(xx[ii-2:ii+1], yy[ii-2:ii+1], ye[ii-2:ii+1])
        new_xx.append(xx[ii])
        inte.append(inte[-1] + inter_i)
        stat_err.append(np.sqrt(stat_err[-1]**2 + inter_se**2))
    return np.array(new_xx), np.array(inte), np.array(stat_err)

def integrate_range_simpson(xx, yy, ye):
    xx = np.array(xx)
    # error esti series 0
    xx0, inte0, stat_err0 = _integrate_range_simpson_inner(xx, yy, ye)
    if len(xx) < 5:
        return xx0, inte0, stat_err0, np.zeros(xx0.shape)
    xx1, inte1, stat_err1 = _integrate_range_simpson_inner(xx[::2], yy[::2], ye[::2])
    diff1 = np.abs(inte1-inte0[::2]) / 16.0
    assert(np.linalg.norm(xx1-xx0[::2]) < 1e-10)
    # error esti series 1, shifted from series 0
    xx2, inte2, stat_err2 = _integrate_range_simpson_inner(xx[2:], yy[2:], ye[2:])
    xx3, inte3, stat_err3 = _integrate_range_simpson_inner(xx[2::2], yy[2::2], ye[2::2])
    from scipy.interpolate import interp1d
    f = interp1d(xx1, diff1)
    diff3 = np.abs(inte3 - inte2[::2]) / 16.0 + f(xx2[0])
    assert(np.linalg.norm(xx3-xx2[::2]) < 1e-10)
    # combine the estimates
    inte_err0 = np.zeros(xx0.shape)
    for ii in range(inte_err0.size):
        if ii % 2 == 0:
            inte_err0[ii] = diff1[ii//2]
        else:
            inte_err0[ii] = diff3[ii//2]    
    return xx0, inte0, inte_err0, stat_err0

def integrate_range (xx, yy, ye, scheme = 's') :
    scheme_ = (scheme.lower()[0])
    if scheme_ == 't':
        return integrate_range_trapezoidal(xx, yy, ye)
    elif scheme_ == 's':
        return integrate_range_simpson(xx, yy, ye)
    else:
        raise RuntimeError('unknow integration scheme', scheme)


def compute_nrefine (all_t, integrand, err, error_scale = None) :
    ntask = all_t.size
    interval_err = []
    interval_err.append(interval_sys_err_trapezoidal(all_t[0:3], integrand[0:3], 'left'))
    for ii in range(1, ntask-2) :
        interval_err.append(
            interval_sys_err_trapezoidal(all_t[ii-1:ii+3], integrand[ii-1:ii+3], 'middle'))
    interval_err.append(interval_sys_err_trapezoidal(all_t[-3:], integrand[-3:], 'right'))
    if error_scale is not None :
        for ii in range(0, ntask-1) :
            interval_err[ii] *= error_scale[ii+1]
    
    interval_nrefine = []
    for ii in range(ntask-1) :
        err_dist = err * (all_t[ii+1] - all_t[ii]) / (all_t[-1] - all_t[0])
        interval_nrefine.append(max(1, int(np.ceil(np.sqrt(interval_err[ii] / err_dist)))))
    #print(interval_nrefine)
    assert(len(interval_nrefine) == len(interval_err))

    return interval_nrefine


def get_task_file_abspath(task_name, file_name): 
    equi_conf = file_name
    cwd = os.getcwd()
    os.chdir(task_name)
    equi_conf = os.path.abspath(equi_conf)
    os.chdir(cwd)
    return equi_conf

def integrate_range_hti(all_lambda, de, de_err, scheme='s'):
    new_lambda, i, i_e, s_e = integrate_range(all_lambda, de, de_err, scheme='s')
    # print('debug:range_hti', new_lambda[-1], all_lambda[-1])
    if new_lambda[-1] != all_lambda[-1] :
        if new_lambda[-1] == all_lambda[-2]:
            _, i1, i_e1, s_e1 = integrate_range(all_lambda[-2:], de[-2:], de_err[-2:], scheme='t')
            diff_e = i[-1] + i1[-1]
            stt_err = np.linalg.norm([s_e[-1], s_e1[-1]])
            sys_err = i_e[-1] + i_e1[-1]
        else :
            raise RuntimeError("lambda does not match!")
    else:
        diff_e = i[-1]
        stt_err = s_e[-1]
        sys_err = i_e[-1]
    return diff_e, stt_err, sys_err


if __name__ == '__main__':
    ninter = 20
    error = 1e-1
    xx = np.arange(0, 1+1e-10, 1./ninter)
    yy = np.exp(xx)
    ye = np.ones(xx.shape) * error / np.sqrt(12)
    yy = yy + (np.random.random([ninter+1]) * error - 0.5 * error)
    print('here', np.random.random([ninter+1]) * error - 0.5 * error)
    # diff0, err0 = integrate_simpson(xx, yy, ye)
    # diff1, err1 = integrate_simpson_nonuniform(xx, yy, ye)
    # real_err0 = np.abs(np.exp(1) - np.exp(0) - diff0)
    # real_err1 = np.abs(np.exp(1) - np.exp(0) - diff1)
    # print('real_error %.2e\nstat_error %.2e' % (real_err0, err0))
    # print('real_error %.2e\nstat_error %.2e' % (real_err1, err1))

    xx1, ii1, i_err1, s_err1 = integrate_range_simpson(xx, yy, ye)
    real_err1 = np.abs(ii1 - (np.exp(xx1) - np.exp(0)))
    print(real_err1)
    print(i_err1)
    print(s_err1)




import os, json, shutil
import numpy as np
import unittest
from unittest.mock import MagicMock, patch, PropertyMock
from context import dpti
from dpti.lib.utils import get_file_md5
from dpti.equi import make_task

class TestEquiMakeTask(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        os.mkdir('tmp_equi/')

    def setUp(self):
        self.maxDiff = None
        self.test_dir = 'tmp_equi'
        self.benchmark_dir = 'benchmark_equi'

    @patch('numpy.random')
    def test_npt(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        test_name = 'npt'
        benchmark_dir = os.path.join(self.benchmark_dir, test_name)
        test_dir = os.path.join(self.test_dir, test_name)

        json_file = os.path.join(benchmark_dir, 'jdata.json')
        with open(json_file) as f:
            jdata = json.load(f)
        dpti.equi.make_task(iter_name=test_dir, jdata=jdata)
        check_file_list = ['in.lammps',  
            'conf.lmp',  'graph.pb']
        for file in check_file_list:
            f1 = os.path.join(benchmark_dir, file)
            f2 = os.path.join(test_dir, file)
            self.assertEqual(get_file_md5(f1), get_file_md5(f2), msg=(f1,f2))

    @patch('numpy.random')
    def test_npt_meam(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        test_name = 'npt_meam'
        benchmark_dir = os.path.join(self.benchmark_dir, test_name)
        test_dir = os.path.join(self.test_dir, test_name)

        json_file = os.path.join(benchmark_dir, 'jdata.json')
        with open(json_file) as f:
            jdata = json.load(f)
        dpti.equi.make_task(iter_name=test_dir, jdata=jdata)
        check_file_list = ['in.lammps', 'conf.lmp', 
            'Sn_18Metals.meam', 'library_18Metals.meam']
        for file in check_file_list:
            f1 = os.path.join(benchmark_dir, file)
            f2 = os.path.join(test_dir, file)
            self.assertEqual(get_file_md5(f1), get_file_md5(f2), msg=(f1,f2))

    @patch('numpy.random')
    def test_nvt(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        test_name = 'nvt'
        benchmark_dir = os.path.join(self.benchmark_dir, test_name)
        test_dir = os.path.join(self.test_dir, test_name)

        json_file = os.path.join(benchmark_dir, 'jdata.json')
        with open(json_file) as f:
            jdata = json.load(f)
        dpti.equi.make_task(iter_name=test_dir, jdata=jdata)
        check_file_list = ['in.lammps',  
            'conf.lmp', 'graph.pb']
        for file in check_file_list:
            f1 = os.path.join(benchmark_dir, file)
            f2 = os.path.join(test_dir, file)
            self.assertEqual(get_file_md5(f1), get_file_md5(f2), msg=(f1,f2))

    @patch('numpy.random')
    def test_nvt_use_npt_avg(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        test_name = 'nvt_use_npt_avg'
        benchmark_dir = os.path.join(self.benchmark_dir, test_name)
        test_dir = os.path.join(self.test_dir, test_name)

        npt_dir = os.path.join('benchmark_equi_log', 'npt')

        json_file = os.path.join(benchmark_dir, 'jdata.json')
        with open(json_file) as f:
            jdata = json.load(f)
        dpti.equi.make_task(iter_name=test_dir, jdata=jdata, npt_dir=npt_dir)

        check_file_list = ['in.lammps',  
            'npt_avg.lmp', 'graph.pb']
        for file in check_file_list:
            f1 = os.path.join(benchmark_dir, file)
            f2 = os.path.join(test_dir, file)
            self.assertEqual(get_file_md5(f1), get_file_md5(f2), msg=(f1,f2))

    @patch('numpy.random')
    def test_water_npt(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        test_name = 'npt_water'

        benchmark_dir = os.path.join(self.benchmark_dir, test_name)
        test_dir = os.path.join(self.test_dir, test_name)


        json_file = os.path.join(benchmark_dir, 'npt.json')
        with open(json_file) as f:
            jdata = json.load(f)
        dpti.equi.make_task(iter_name=test_dir, jdata=jdata)

        check_file_list = ['in.lammps']
        for file in check_file_list:
            f1 = os.path.join(benchmark_dir, file)
            f2 = os.path.join(test_dir, file)
            self.assertEqual(get_file_md5(f1), get_file_md5(f2), msg=(f1,f2))

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree('tmp_equi/')

if __name__ == '__main__':
    unittest.main()



from airflow import DAG
from dpti.dags.dp_ti_gdi import GDIDAGFactory

GDI_dag_factory = GDIDAGFactory(gdi_name='Sn_gdi_270K_test31', 
    dag_work_base='/home/fengbo/4_Sn/16_gdi_3/test_Sn.p.270K.1GPa-2GPa')
loop_dag = GDI_dag_factory.loop_dag
main_dag = GDI_dag_factory.main_dag





import os, textwrap
import numpy as np
import unittest
from context import dpti
from potential_common import soft_param, soft_param_three_element, meam_model
from dpti.hti_liq import _ff_deep_on

class TestDeepOn(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_one_element(self):
        input = dict(lamb=0.075,
            sparam=soft_param, model="graph.pb",
            if_meam=False, meam_model=None)
        ret1 = textwrap.dedent("""\
        variable        EPSILON equal 0.030000
        variable        ONE equal 1
        pair_style      hybrid/overlay deepmd graph.pb lj/cut/soft 1.000000 0.500000 6.000000
        pair_coeff      * * deepmd
        pair_coeff      1 1 lj/cut/soft ${EPSILON} 2.493672 0.500000
        fix             tot_pot all adapt/fep 0 pair deepmd scale * * v_LAMBDA
        compute         e_diff all fep ${TEMP} pair deepmd scale * * v_ONE
        """)
        ret2 = _ff_deep_on(**input)
        self.assertEqual(ret1, ret2)
    
    def test_three_element(self):
        input = dict(lamb=0.075,
            sparam=soft_param_three_element, model="graph.pb",
            if_meam=False, meam_model=None)

        ret1 = textwrap.dedent("""\
        variable        EPSILON equal 0.030000
        variable        ONE equal 1
        pair_style      hybrid/overlay deepmd graph.pb lj/cut/soft 1.000000 0.600000 6.000000
        pair_coeff      * * deepmd
        pair_coeff      1 1 lj/cut/soft ${EPSILON} 2.000000 0.500000
        pair_coeff      1 2 lj/cut/soft ${EPSILON} 2.010000 0.500000
        pair_coeff      1 3 lj/cut/soft ${EPSILON} 2.020000 0.500000
        pair_coeff      2 2 lj/cut/soft ${EPSILON} 2.110000 0.500000
        pair_coeff      2 3 lj/cut/soft ${EPSILON} 2.120000 0.500000
        pair_coeff      3 3 lj/cut/soft ${EPSILON} 2.220000 0.500000
        fix             tot_pot all adapt/fep 0 pair deepmd scale * * v_LAMBDA
        compute         e_diff all fep ${TEMP} pair deepmd scale * * v_ONE
        """)
        
        ret2 = _ff_deep_on(**input)
        self.assertEqual(ret1, ret2)

    def test_deepmd(self):
        input = dict(lamb=0.075,
            sparam=soft_param, model="graph.pb",
            if_meam=False, meam_model=None)
        ret1 = textwrap.dedent("""\
        variable        EPSILON equal 0.030000
        variable        ONE equal 1
        pair_style      hybrid/overlay deepmd graph.pb lj/cut/soft 1.000000 0.500000 6.000000
        pair_coeff      * * deepmd
        pair_coeff      1 1 lj/cut/soft ${EPSILON} 2.493672 0.500000
        fix             tot_pot all adapt/fep 0 pair deepmd scale * * v_LAMBDA
        compute         e_diff all fep ${TEMP} pair deepmd scale * * v_ONE
        """)
        ret2 = _ff_deep_on(**input)
        self.assertEqual(ret1, ret2)
        # print('---------')
        # print(ret1)
        # print('---------')
        # print(ret2)
        # print('---------')

        # print(ret2)
    
    def test_meam(self):
        input = dict(lamb=0.075,
            sparam=soft_param, model=None,
            if_meam=True, meam_model=meam_model)
        ret1 = textwrap.dedent("""\
        variable        EPSILON equal 0.030000
        variable        ONE equal 1
        pair_style      hybrid/overlay meam lj/cut/soft 1.000000 0.500000 6.000000
        pair_coeff      * * meam library_18Metals.meam Sn Sn_18Metals.meam Sn
        pair_coeff      1 1 lj/cut/soft ${EPSILON} 2.493672 0.500000
        fix             tot_pot all adapt/fep 0 pair meam scale * * v_LAMBDA
        compute         e_diff all fep ${TEMP} pair meam scale * * v_ONE
        """)
        ret2 = _ff_deep_on(**input)
        self.assertEqual(ret1, ret2)
        # print('---------')
        # print(ret1)
        # print('---------')
        # print(ret2)
        # print('---------')



    # def test_three_element(self):
    #     input = dict(lamb=0.075, sparam=soft_param_three_element)
    #     ret1 = textwrap.dedent("""\
    #     variable        EPSILON equal 0.030000
    #     pair_style      lj/cut/soft 1.000000 0.600000 6.000000
    #     pair_coeff      1 1 ${EPSILON} 2.000000 0.500000
    #     pair_coeff      1 2 ${EPSILON} 2.010000 0.500000
    #     pair_coeff      1 3 ${EPSILON} 2.020000 0.500000
    #     pair_coeff      2 2 ${EPSILON} 2.110000 0.500000
    #     pair_coeff      2 3 ${EPSILON} 2.120000 0.500000
    #     pair_coeff      3 3 ${EPSILON} 2.220000 0.500000
    #     fix             tot_pot all adapt/fep 0 pair lj/cut/soft epsilon * * v_LAMBDA scale yes
    #     compute         e_diff all fep ${TEMP} pair lj/cut/soft epsilon * * v_EPSILON
    #     """)
    #     ret2 = _ff_soft_on(**input)
    #     self.assertEqual(ret1, ret2)
        # print(ret2)
        # pass
        # input = dict(lamb=0.075, m_spring_k=[118.71],
        #     var_spring=False)
        # ret1 = textwrap.dedent("""\
        # group           type_1 type 1
        # fix             l_spring_1 type_1 spring/self 1.1871000000e+02
        # fix_modify      l_spring_1 energy yes
        # variable        l_spring equal f_l_spring_1
        # """)
        # ret2 = dpti.hti._ff_spring(**input)
        # self.assertEqual(ret1, ret2)


#!/usr/bin/env python3

import os, sys, paramiko, json, uuid, tarfile, time, stat
from enum import Enum

class JobStatus (Enum) :
    unsubmitted = 1
    waiting = 2
    running = 3
    terminated = 4
    finished = 5
    unknow = 100

def _default_item(resources, key, value) :
    if key not in resources :
        resources[key] = value

def _set_default_resource(res) :
    if res == None :
        res = {}
    _default_item(res, 'numb_node', 1)
    _default_item(res, 'task_per_node', 1)
    _default_item(res, 'numb_gpu', 0)
    _default_item(res, 'time_limit', '1:0:0')
    _default_item(res, 'mem_limit', -1)
    _default_item(res, 'partition', '')
    _default_item(res, 'account', '')
    _default_item(res, 'qos', '')
    _default_item(res, 'constraint_list', [])
    _default_item(res, 'license_list', [])
    _default_item(res, 'exclude_list', [])
    _default_item(res, 'module_unload_list', [])
    _default_item(res, 'module_list', [])
    _default_item(res, 'source_list', [])
    _default_item(res, 'envs', None)
    _default_item(res, 'with_mpi', False)


class SSHSession (object) :
    def __init__ (self, jdata) :
        self.remote_profile = jdata
        # with open(remote_profile) as fp :
        #     self.remote_profile = json.load(fp)
        self.remote_host = self.remote_profile['hostname']
        self.remote_port = self.remote_profile['port']
        self.remote_uname = self.remote_profile['username']
        self.remote_workpath = self.remote_profile['work_path']
        self.ssh = self._setup_ssh(self.remote_host, self.remote_port, username = self.remote_uname)
        
    def _setup_ssh(self,
                   hostname,
                   port, 
                   username = None,
                   password = None):
        ssh_client = paramiko.SSHClient()        
        ssh_client.load_system_host_keys()
        ssh_client.set_missing_host_key_policy(paramiko.WarningPolicy)
        ssh_client.connect(hostname, port=port, username=username, password=password)
        assert(ssh_client.get_transport().is_active())
        return ssh_client

    def get_ssh_client(self) :
        return self.ssh

    def get_session_root(self) :
        return self.remote_workpath

    def close(self) :
        self.ssh.close()


class RemoteJob (object):
    def __init__ (self,
                  ssh_session,
                  local_root
    ) :
        
        self.local_root = os.path.abspath(local_root)
        self.job_uuid = str(uuid.uuid4())
        # self.job_uuid = 'a21d0017-c9f1-4d29-9a03-97df06965cef'
        self.remote_root = os.path.join(ssh_session.get_session_root(), self.job_uuid)
        print("local_root is ", local_root)
        print("remote_root is", self.remote_root)
        self.ssh = ssh_session.get_ssh_client()        
        sftp = self.ssh.open_sftp()        
        sftp.mkdir(self.remote_root)
        sftp.close()
        # open('job_uuid', 'w').write(self.job_uuid)
        
    def get_job_root(self) :
        return self.remote_root
        
    def upload(self,
               job_dirs,
               local_up_files,
               dereference = True) :
        cwd = os.getcwd()
        os.chdir(self.local_root) 
        file_list = []
        for ii in job_dirs :
            for jj in local_up_files :
                file_list.append(os.path.join(ii,jj))        
        self._put_files(file_list, dereference = dereference)
        os.chdir(cwd)

    def download(self, 
                 job_dirs,
                 remote_down_files) :
        cwd = os.getcwd()
        os.chdir(self.local_root) 
        file_list = []
        for ii in job_dirs :
            for jj in remote_down_files :
                file_list.append(os.path.join(ii,jj))
        self._get_files(file_list)
        os.chdir(cwd)
        
    def block_checkcall(self, 
                        cmd) :
        stdin, stdout, stderr = self.ssh.exec_command(('cd %s ;' % self.remote_root) + cmd)
        exit_status = stdout.channel.recv_exit_status() 
        if exit_status != 0:
            raise RuntimeError("Get error code %d in calling through ssh with job: %s ", (exit_status, self.job_uuid))
        return stdin, stdout, stderr    

    def block_call(self, 
                   cmd) :
        stdin, stdout, stderr = self.ssh.exec_command(('cd %s ;' % self.remote_root) + cmd)
        exit_status = stdout.channel.recv_exit_status() 
        return exit_status, stdin, stdout, stderr

    def clean(self) :        
        sftp = self.ssh.open_sftp()        
        self._rmtree(sftp, self.remote_root)
        sftp.close()

    def _rmtree(self, sftp, remotepath, level=0, verbose = False):
        for f in sftp.listdir_attr(remotepath):
            rpath = os.path.join(remotepath, f.filename)
            if stat.S_ISDIR(f.st_mode):
                self._rmtree(sftp, rpath, level=(level + 1))
            else:
                rpath = os.path.join(remotepath, f.filename)
                if verbose: print('removing %s%s' % ('    ' * level, rpath))
                sftp.remove(rpath)
        if verbose: print('removing %s%s' % ('    ' * level, remotepath))
        sftp.rmdir(remotepath)

    def _put_files(self,
                   files,
                   dereference = True) :
        of = self.job_uuid + '.tgz'
        # local tar
        cwd = os.getcwd()
        os.chdir(self.local_root)
        if os.path.isfile(of) :
            os.remove(of)
        with tarfile.open(of, "w:gz", dereference = dereference) as tar:
            for ii in files :
                tar.add(ii)
        os.chdir(cwd)
        # trans
        from_f = os.path.join(self.local_root, of)
        to_f = os.path.join(self.remote_root, of)
        sftp = self.ssh.open_sftp()
        sftp.put(from_f, to_f)
        # remote extract
        self.block_checkcall('tar xf %s' % of)
        # clean up
        os.remove(from_f)
        sftp.remove(to_f)
        sftp.close()

    def _get_files(self, 
                   files) :
        of = self.job_uuid + '.tgz'
        flist = ""
        for ii in files :
            flist += " " + ii
        # remote tar
        self.block_checkcall('tar czf %s %s' % (of, flist))
        # trans
        from_f = os.path.join(self.remote_root, of)
        to_f = os.path.join(self.local_root, of)
        if os.path.isfile(to_f) :
            os.remove(to_f)
        sftp = self.ssh.open_sftp()
        sftp.get(from_f, to_f)
        # extract
        cwd = os.getcwd()
        os.chdir(self.local_root)
        with tarfile.open(of, "r:gz") as tar:
            tar.extractall()
        os.chdir(cwd)        
        # cleanup
        os.remove(to_f)
        sftp.remove(from_f)

class CloudMachineJob (RemoteJob) :
    def submit(self, 
               job_dirs,
               cmd, 
               args = None, 
               resources = None) :
        
        #print("Current path is",os.getcwd())

        #for ii in job_dirs :
        #    if not os.path.isdir(ii) :
        #        raise RuntimeError("cannot find dir %s" % ii)
        # print(self.remote_root)
        script_name = self._make_script(job_dirs, cmd, args, resources)
        self.stdin, self.stdout, self.stderr = self.ssh.exec_command(('cd %s; bash %s' % (self.remote_root, script_name)))
        # print(self.stderr.read().decode('utf-8'))
        # print(self.stdout.read().decode('utf-8'))

    def check_status(self) :
        if not self._check_finish(self.stdout) :
            return JobStatus.running
        elif self._get_exit_status(self.stdout) == 0 :
            return JobStatus.finished
        else :
            return JobStatus.terminated

    def _check_finish(self, stdout) :
        return stdout.channel.exit_status_ready()
    
    def _get_exit_status(self, stdout) :
        return stdout.channel.recv_exit_status() 
    
    def _make_script(self, 
                     job_dirs,
                     cmd, 
                     args = None, 
                     resources = None) :
        _set_default_resource(resources)
        envs = resources['envs']
        module_list = resources['module_list']
        module_unload_list = resources['module_unload_list']
        task_per_node = resources['task_per_node']

        script_name = 'run.sh'
        if args == None :
            args = []
            for ii in job_dirs:
                args.append('')
        script = os.path.join(self.remote_root, script_name)
        sftp = self.ssh.open_sftp()
        with sftp.open(script, 'w') as fp :
            fp.write('#!/bin/bash\n\n')
            # fp.write('set -euo pipefail\n')
            if envs != None :
                for key in envs.keys() :
                    fp.write('export %s=%s\n' % (key, envs[key]))
                fp.write('\n')
            if module_unload_list is not None :
                for ii in module_unload_list :
                    fp.write('module unload %s\n' % ii)
                fp.write('\n')
            if module_list is not None :
                for ii in module_list :
                    fp.write('module load %s\n' % ii)
                fp.write('\n')
            for ii,jj in zip(job_dirs, args) :
                fp.write('cd %s\n' % ii)                
                fp.write('test $? -ne 0 && exit\n')
                if resources['with_mpi'] == True :
                    fp.write('mpirun -n %d %s %s\n' 
                             % (task_per_node, cmd, jj))
                else :
                    fp.write('%s %s\n' % (cmd, jj))
                fp.write('test $? -ne 0 && exit\n')
                fp.write('cd %s\n' % self.remote_root)         
                fp.write('test $? -ne 0 && exit\n')  
            fp.write('\ntouch tag_finished\n')
        sftp.close()
        return script_name


class SlurmJob (RemoteJob) :
    def submit(self, 
               job_dirs,
               cmd,
               args = None, 
               resources = None) :
        script_name = self._make_script(job_dirs, cmd, args, res = resources)
        stdin, stdout, stderr = self.block_checkcall(('cd %s; sbatch %s' % (self.remote_root, script_name)))
        subret = (stdout.readlines())
        job_id = subret[0].split()[-1]
        sftp = self.ssh.open_sftp()
        with sftp.open(os.path.join(self.remote_root, 'job_id'), 'w') as fp:
            fp.write(job_id)
        sftp.close()

    def check_status(self) :
        job_id = self._get_job_id()
        if job_id == "" :
            raise RuntimeError("job %s is has not been submitted" % self.remote_root)
        ret, stdin, stdout, stderr\
            = self.block_call ("squeue --job " + job_id)
        err_str = stderr.read().decode('utf-8')
        if (ret != 0) :
            if str("Invalid job id specified") in err_str :
                if self._check_finish_tag() :
                    return JobStatus.finished
                else :
                    return JobStatus.terminated
            else :
                raise RuntimeError\
                    ("status command squeue fails to execute\nerror message:%s\nreturn code %d\n" % (err_str, ret))
        status_line = stdout.read().decode('utf-8').split ('\n')[-2]
        status_word = status_line.split ()[-4]
        if      status_word in ["PD","CF","S"] :
            return JobStatus.waiting
        elif    status_word in ["R","CG"] :
            return JobStatus.running
        elif    status_word in ["C","E","K","BF","CA","CD","F","NF","PR","SE","ST","TO"] :
            if self._check_finish_tag() :
                return JobStatus.finished
            else :
                return JobStatus.terminated
        else :
            return JobStatus.unknown
    
    def _get_job_id(self) :
        sftp = self.ssh.open_sftp()
        with sftp.open(os.path.join(self.remote_root, 'job_id'), 'r') as fp:            
            ret = fp.read().decode('utf-8')
        sftp.close()
        return ret

    def _check_finish_tag(self) :
        sftp = self.ssh.open_sftp()
        try:
            sftp.stat(os.path.join(self.remote_root, 'tag_finished')) 
            ret =  True
        except IOError:
            ret = False
        sftp.close()
        return ret

    def _make_script(self, 
                     job_dirs,
                     cmd,
                     args = None, 
                     res = None) :
        _set_default_resource(res)
        ret = ''
        ret += "#!/bin/bash -l\n"
        ret += "#SBATCH -N %d\n" % res['numb_node']
        ret += "#SBATCH --ntasks-per-node %d\n" % res['task_per_node']
        ret += "#SBATCH -t %s\n" % res['time_limit']
        if res['mem_limit'] > 0 :
            ret += "#SBATCH --mem %dG \n" % res['mem_limit']
        if len(res['account']) > 0 :
            ret += "#SBATCH --account %s \n" % res['account']
        if len(res['partition']) > 0 :
            ret += "#SBATCH --partition %s \n" % res['partition']
        if len(res['qos']) > 0 :
            ret += "#SBATCH --qos %s \n" % res['qos']
        if res['numb_gpu'] > 0 :
            ret += "#SBATCH --gres=gpu:%d\n" % res['numb_gpu']
        for ii in res['constraint_list'] :
            ret += '#SBATCH -C %s \n' % ii
        for ii in res['license_list'] :
            ret += '#SBATCH -L %s \n' % ii
        for ii in res['exclude_list'] :
            ret += '#SBATCH --exclude %s \n' % ii
        ret += "\n"
        # ret += 'set -euo pipefail\n\n'
        for ii in res['module_unload_list'] :
            ret += "module unload %s\n" % ii
        for ii in res['module_list'] :
            ret += "module load %s\n" % ii
        ret += "\n"
        for ii in res['source_list'] :
            ret += "source %s\n" %ii
        ret += "\n"
        envs = res['envs']
        if envs != None :
            for key in envs.keys() :
                ret += 'export %s=%s\n' % (key, envs[key])
            ret += '\n'

        if args == None :
            args = []
            for ii in job_dirs:
                args.append('')
        for ii,jj in zip(job_dirs, args) :
            ret += 'cd %s\n' % ii
            ret += 'test $? -ne 0 && exit\n'
            if res['with_mpi'] :
                ret += 'srun %s %s\n' % (cmd, jj)
            else :
                ret += '%s %s\n' % (cmd, jj)
            ret += 'test $? -ne 0 && exit\n'
            ret += 'cd %s\n' % self.remote_root
            ret += 'test $? -ne 0 && exit\n'
        ret += '\ntouch tag_finished\n'

        script_name = 'run.sub'
        script = os.path.join(self.remote_root, script_name)
        sftp = self.ssh.open_sftp()
        with sftp.open(script, 'w') as fp :
            fp.write(ret)
        sftp.close()

        return script_name


class PBSJob (RemoteJob) :
    def submit(self, 
               job_dirs,
               cmd,
               args = None, 
               resources = None) :
        script_name = self._make_script(job_dirs, cmd, args, res = resources)
        stdin, stdout, stderr = self.block_checkcall(('cd %s; qsub %s' % (self.remote_root, script_name)))
        subret = (stdout.readlines())
        job_id = subret[0].split()[0]
        sftp = self.ssh.open_sftp()
        with sftp.open(os.path.join(self.remote_root, 'job_id'), 'w') as fp:
            fp.write(job_id)
        sftp.close()

    def check_status(self) :
        job_id = self._get_job_id()
        if job_id == "" :
            raise RuntimeError("job %s is has not been submitted" % self.remote_root)
        ret, stdin, stdout, stderr\
            = self.block_call ("qstat -x " + job_id)
        err_str = stderr.read().decode('utf-8')
        if (ret != 0) :
            if str("qstat: Unknown Job Id") in err_str :
                if self._check_finish_tag() :
                    return JobStatus.finished
                else :
                    return JobStatus.terminated
            else :
                raise RuntimeError ("status command qstat fails to execute. erro info: %s return code %d"
                                    % (err_str, ret))
        status_line = stdout.read().decode('utf-8').split ('\n')[-2]
        status_word = status_line.split ()[-2]        
#        print (status_word)
        if      status_word in ["Q","H"] :
            return JobStatus.waiting
        elif    status_word in ["R"] :
            return JobStatus.running
        elif    status_word in ["C","E","K","F"] :
            if self._check_finish_tag() :
                return JobStatus.finished
            else :
                return JobStatus.terminated
        else :
            return JobStatus.unknown
    
    def _get_job_id(self) :
        sftp = self.ssh.open_sftp()
        with sftp.open(os.path.join(self.remote_root, 'job_id'), 'r') as fp:
            ret = fp.read().decode('utf-8')
        sftp.close()
        return ret

    def _check_finish_tag(self) :
        sftp = self.ssh.open_sftp()
        try:
            sftp.stat(os.path.join(self.remote_root, 'tag_finished')) 
            ret =  True
        except IOError:
            ret = False
        sftp.close()
        return ret

    def _make_script(self, 
                     job_dirs,
                     cmd,
                     args = None, 
                     res = None) :
        _set_default_resource(res)
        ret = ''
        ret += "#!/bin/bash -l\n"
        if res.get('hpc_job_name', None):
            ret += '#PBS -N %s\n' % res['hpc_job_name']
        if res['numb_gpu'] == 0:
            ret += '#PBS -l select=%d:ncpus=%d\n' % (res['numb_node'], res['task_per_node'])
        else :
            ret += '#PBS -l select=%d:ncpus=%d:ngpus=%d\n' % (res['numb_node'], res['task_per_node'], res['numb_gpu'])
        ret += '#PBS -l walltime=%s\n' % (res['time_limit'])
        # if res['mem_limit'] > 0 :
        #     ret += "#PBS -l mem=%dG \n" % res['mem_limit']
        ret += '#PBS -j oe\n'
        if len(res['partition']) > 0 :
            ret += '#PBS -q %s\n' % res['partition']
        ret += "\n"
        for ii in res['module_unload_list'] :
            ret += "module unload %s\n" % ii
        for ii in res['module_list'] :
            ret += "module load %s\n" % ii
        ret += "\n"
        for ii in res['source_list'] :
            ret += "source %s\n" %ii
        ret += "\n"
        envs = res['envs']
        if envs != None :
            for key in envs.keys() :
                ret += 'export %s=%s\n' % (key, envs[key])
            ret += '\n'
        ret += 'cd $PBS_O_WORKDIR\n\n'

        if args == None :
            args = []
            for ii in job_dirs:
                args.append('')
        for ii,jj in zip(job_dirs, args) :
            ret += 'cd %s\n' % ii
            ret += 'test $? -ne 0 && exit\n'
            if res['with_mpi'] :
                ret += 'mpirun -machinefile $PBS_NODEFILE -n %d %s %s\n' % (res['numb_node'] * res['task_per_node'], cmd, jj)
            else :
                ret += '%s %s\n' % (cmd, jj)                
            ret += 'test $? -ne 0 && exit\n'
            ret += 'cd %s\n' % self.remote_root
            ret += 'test $? -ne 0 && exit\n'
        ret += '\ntouch tag_finished\n'

        script_name = 'run.sub'
        script = os.path.join(self.remote_root, script_name)
        sftp = self.ssh.open_sftp()
        with sftp.open(script, 'w') as fp :
            fp.write(ret)
        sftp.close()

        return script_name


# ssh_session = SSHSession('localhost.json')        
# rjob = CloudMachineJob(ssh_session, '.')
# # can upload dirs and normal files
# rjob.upload(['job0', 'job1'], ['batch_exec.py', 'test'])
# rjob.submit(['job0', 'job1'], 'touch a; sleep 2')
# while rjob.check_status() == JobStatus.running :
#     print('checked')
#     time.sleep(2)
# print(rjob.check_status())
# # can download dirs and normal files
# rjob.download(['job0', 'job1'], ['a'])
# # rjob.clean()



#!/usr/bin/env python3

import os, sys, json, argparse, glob
import numpy as np
import scipy.constants as pc

# sys.path.insert(0, os.path.join(os.path.dirname(__file__), '../'))
import dpti
from dpti.lib.utils import create_dict_not_empty_key, create_path, relative_link_file
from dpti.lib.utils import block_avg, link_file_in_dict
from dpti.lib.water import compute_bonds
from dpti.lib.water import posi_diff
from dpti.lib.utils import get_task_file_abspath
# import dpti

from dpti.lib.lammps import get_natoms, get_last_dump, get_thermo
from dpti.lib.lmp import from_system_data
from dpti.lib.dump import system_data

from dargs import dargs, Argument, Variant
# from lib import dump 
# from .lib import lammps
# from .lib import dump
# from .lib import lmp

# def _gen_lammps_input
# def gen_equi_header(nsteps, prt_freq, dump_freq, temp, pres, tau_t, tau_p, mass_map, conf_file):
# def gen_equi_header(nsteps, prt_freq, dump_freq, temp, pres, tau_t, tau_p, mass_map, conf_file):
def gen_equi_header(nsteps, 
        thermo_freq, 
        dump_freq,
        mass_map,
        temp,
        tau_t,
        tau_p,
        equi_conf,
        pres=None
    ):
    ret = ''
    ret += 'clear\n'
    ret += '# --------------------- VARIABLES-------------------------\n'
    ret += 'variable        NSTEPS          equal %d\n' % nsteps
    ret += 'variable        THERMO_FREQ     equal %d\n' % thermo_freq
    ret += 'variable        DUMP_FREQ       equal %d\n' % dump_freq
    ret += 'variable        NREPEAT         equal ${NSTEPS}/${DUMP_FREQ}\n'
    ret += 'variable        TEMP            equal %.6f\n' % temp
    # if equi_settings['pres'] is not None :
    if pres is not None:
        ret += 'variable        PRES            equal %.6f\n' % pres
    ret += 'variable        TAU_T           equal %.6f\n' % tau_t
    ret += 'variable        TAU_P           equal %.6f\n' % tau_p
    ret += '# ---------------------- INITIALIZAITION ------------------\n'
    ret += 'units           metal\n'
    ret += 'boundary        p p p\n'
    ret += 'atom_style      atomic\n'
    ret += '# --------------------- ATOM DEFINITION ------------------\n'
    ret += 'box             tilt large\n'
    ret += 'read_data       %s\n' % equi_conf
    ret += 'change_box      all triclinic\n'
    for jj in range(len(mass_map)):
        ret+= "mass            %d %.6f\n" %(jj+1, mass_map[jj])
    return ret
    
# def gen_equi_force_field(model, if_meam=None):
def gen_equi_force_field(model, if_meam=False, meam_model=None):
    # equi_settings =
    # model = equi_settings['model']
    # assert type(model) is dict, f"equi_settings['model] must be a dict. model:{model}"
    # model_type = equi_settings.get('model_type', None)
    # if model_type == 'deepmd':
    #     assert 'deepmd_model' in equi_settings, f" 'deepmd_model' must be in equi_settings. {equi_settings}"
    # elif model_type == 'meam':
    #     assert 'meam_library' in equi_settings, f" 'meam_library' must be in equi_settings. {equi_settings}"
    #     assert 'meam_potential' in equi_settings, f" 'meam_potential' must be in equi_settings. {equi_settings}"
    #     assert 'meam_element' in equi_settings, f" 'meam_element' must be in equi_settings. {equi_settings}"
    # else: 
    #     raise ValueError(f" model_type:{model_type} must be in ['deepmd', 'meam'];"
    #         f"equi_settings:{equi_settings}")

    ret = ''
    ret += '# --------------------- FORCE FIELDS ---------------------\n'
    if not if_meam:
        ret += 'pair_style      deepmd %s\n' % model
        ret += 'pair_coeff\n'
    else:
        meam_library = meam_model['library']
        meam_potential = meam_model['potential']
        meam_element = meam_model['element']
        ret += 'pair_style      meam\n'
        ret += 'pair_coeff      * * %s %s %s %s\n' % (meam_library,
            meam_element, meam_potential, meam_element)
    return ret

def gen_equi_thermo_settings(timestep):
    ret = ''
    ret += '# --------------------- MD SETTINGS ----------------------\n'    
    ret += 'neighbor        1.0 bin\n'
    ret += 'timestep        %.6f\n' % timestep
    ret += 'thermo          ${THERMO_FREQ}\n'
    ret += 'compute         allmsd all msd\n'
    ret += 'thermo_style    custom step ke pe etotal enthalpy temp press vol lx ly lz xy xz yz pxx pyy pzz pxy pxz pyz c_allmsd[*]\n'
    return ret

def gen_equi_dump_settings(if_dump_avg_posi):
    ret = ''
    if if_dump_avg_posi:
        ret += 'compute         ru all property/atom xu yu zu\n'
        ret += 'fix             ap all ave/atom ${DUMP_FREQ} ${NREPEAT} ${NSTEPS} c_ru[1] c_ru[2] c_ru[3]\n'
        ret += 'dump            fp all custom ${NSTEPS} dump.avgposi id type f_ap[1] f_ap[2] f_ap[3]\n'
    ret += 'dump            1 all custom ${DUMP_FREQ} dump.equi id type x y z vx vy vz\n'
    return ret

def gen_equi_ensemble_settings(ens):
    # ens = equi_settings['ens']
    ret = ''
    if ens == 'nvt' :
        ret += 'fix             1 all nvt temp ${TEMP} ${TEMP} ${TAU_T}\n'
    elif ens == 'npt-iso' or ens == 'npt':
        ret += 'fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}\n'
    elif ens == 'npt-xy' :
        ret += 'fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} aniso ${PRES} ${PRES} ${TAU_P} couple xy\n'
    elif ens == 'npt-aniso' :
        ret += 'fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} aniso ${PRES} ${PRES} ${TAU_P}\n'
    elif ens == 'npt-tri' :
        ret += 'fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} tri ${PRES} ${PRES} ${TAU_P}\n'
    elif ens == 'nve' :
        ret += 'fix             1 all nve\n'
    else :
        raise RuntimeError('unknow ensemble %s\n' % ens)        
    ret += 'fix             mzero all momentum 10 linear 1 1 1\n'
    ret += '# --------------------- INITIALIZE -----------------------\n'    
    ret += 'velocity        all create ${TEMP} %d\n' % (np.random.randint(1, 2**16))
    ret += 'velocity        all zero linear\n'
    ret += '# --------------------- RUN ------------------------------\n'    
    ret += 'run             ${NSTEPS}\n'
    ret += 'write_data      out.lmp\n'
    return ret

def gen_equi_lammps_input(nsteps, 
        thermo_freq, 
        dump_freq,
        mass_map,
        temp,
        tau_t,
        tau_p,
        equi_conf,
        model,
        timestep,
        if_dump_avg_posi,
        ens,
        pres=None,
        if_meam=False, 
        meam_model=None
    ):
    equi_header = gen_equi_header(nsteps, 
        thermo_freq=thermo_freq, 
        dump_freq=dump_freq,
        mass_map=mass_map,
        temp=temp,
        tau_t=tau_t,
        tau_p=tau_p,
        equi_conf=equi_conf,
        pres=pres
    )
    equi_force_field = gen_equi_force_field(
        model, if_meam=if_meam, meam_model=meam_model
    )
    equi_thermo_settings = gen_equi_thermo_settings(
        timestep=timestep
    )
    equi_dump_settings = gen_equi_dump_settings(
        if_dump_avg_posi=if_dump_avg_posi
    )
    equi_ensemble_settings = gen_equi_ensemble_settings(
        ens=ens
    )

    equi_lammps_input = (equi_header + equi_force_field
        + equi_thermo_settings + equi_dump_settings + equi_ensemble_settings)
    return equi_lammps_input
# def 

# def _gen_lammps_input (conf_file, 
#                        mass_map,
#                        model,
#                        nsteps,
#                        dt,
#                        ens,
#                        temp,
#                        pres = 1.0, 
#                        tau_t = 0.1,
#                        tau_p = 0.5,
#                        prt_freq = 100, 
#                        dump_freq = 1000, 
#                        dump_ave_posi = False,
#                        if_meam = False) :
#     ret = ''
#     ret += 'clear\n'
#     ret += '# --------------------- VARIABLES-------------------------\n'
#     ret += 'variable        NSTEPS          equal %d\n' % nsteps
#     ret += 'variable        THERMO_FREQ     equal %d\n' % prt_freq
#     ret += 'variable        DUMP_FREQ       equal %d\n' % dump_freq
#     ret += 'variable        NREPEAT         equal ${NSTEPS}/${DUMP_FREQ}\n'
#     ret += 'variable        TEMP            equal %f\n' % temp
#     if pres is not None :
#         ret += 'variable        PRES            equal %f\n' % pres
#     ret += 'variable        TAU_T           equal %f\n' % tau_t
#     ret += 'variable        TAU_P           equal %f\n' % tau_p
#     ret += '# ---------------------- INITIALIZAITION ------------------\n'
#     ret += 'units           metal\n'
#     ret += 'boundary        p p p\n'
#     ret += 'atom_style      atomic\n'
#     ret += '# --------------------- ATOM DEFINITION ------------------\n'
#     ret += 'box             tilt large\n'
#     ret += 'read_data       %s\n' % conf_file
#     ret += 'change_box      all triclinic\n'
#     for jj in range(len(mass_map)) :
#         ret+= "mass            %d %f\n" %(jj+1, mass_map[jj])
#     ret += '# --------------------- FORCE FIELDS ---------------------\n'
#     if if_meam:
#         ret += 'pair_style      meam\n'
#         ret += 'pair_coeff      * * /home/fengbo/4_Sn/meam_files/library_18Metal.meam Sn /home/fengbo/4_Sn/meam_files/Sn_18Metal.meam Sn\n'
#     else:
#         ret += 'pair_style      deepmd %s\n' % model
#         ret += 'pair_coeff\n'
#     ret += '# --------------------- MD SETTINGS ----------------------\n'    
#     ret += 'neighbor        1.0 bin\n'
#     ret += 'timestep        %s\n' % dt
#     ret += 'thermo          ${THERMO_FREQ}\n'
#     ret += 'compute         allmsd all msd\n'
#     if ens == 'nvt' :        
#         ret += 'thermo_style    custom step ke pe etotal enthalpy temp press vol lx ly lz xy xz yz pxx pyy pzz pxy pxz pyz c_allmsd[*]\n'
#     elif 'npt' in ens :
#         ret += 'thermo_style    custom step ke pe etotal enthalpy temp press vol lx ly lz xy xz yz pxx pyy pzz pxy pxz pyz c_allmsd[*]\n'
#     else :	
#         raise RuntimeError('unknow ensemble %s\n' % ens)
#     if dump_ave_posi: 
#         ret += 'compute         ru all property/atom xu yu zu\n'
#         ret += 'fix             ap all ave/atom ${DUMP_FREQ} ${NREPEAT} ${NSTEPS} c_ru[1] c_ru[2] c_ru[3]\n'
#         ret += 'dump            fp all custom ${NSTEPS} dump.avgposi id type f_ap[1] f_ap[2] f_ap[3]\n'
#     ret += 'dump            1 all custom ${DUMP_FREQ} dump.equi id type x y z vx vy vz\n'
#     if ens == 'nvt' :
#         ret += 'fix             1 all nvt temp ${TEMP} ${TEMP} ${TAU_T}\n'
#     elif ens == 'npt-iso' or ens == 'npt':
#         ret += 'fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}\n'
#     elif ens == 'npt-xy' :
#         ret += 'fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} aniso ${PRES} ${PRES} ${TAU_P} couple xy\n'
#     elif ens == 'npt-aniso' :
#         ret += 'fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} aniso ${PRES} ${PRES} ${TAU_P}\n'
#     elif ens == 'npt-tri' :
#         ret += 'fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} tri ${PRES} ${PRES} ${TAU_P}\n'
#     elif ens == 'nve' :
#         ret += 'fix             1 all nve\n'
#     else :
#         raise RuntimeError('unknow ensemble %s\n' % ens)        
#     ret += 'fix             mzero all momentum 10 linear 1 1 1\n'
#     ret += '# --------------------- INITIALIZE -----------------------\n'    
#     ret += 'velocity        all create ${TEMP} %d\n' % (np.random.randint(1, 2**16))
#     ret += 'velocity        all zero linear\n'
#     ret += '# --------------------- RUN ------------------------------\n'    
#     ret += 'run             ${NSTEPS}\n'
#     ret += 'write_data      out.lmp\n'
#     return ret

def npt_equi_conf(npt_dir) :
    thermo_file = os.path.join(npt_dir, 'log.lammps')
    dump_file = os.path.join(npt_dir, 'dump.equi')
    j_file = os.path.join(npt_dir, 'equi_settings.json')
    with open(j_file, 'r') as f:
        jdata = json.load(f)
    stat_skip = jdata['stat_skip']
    stat_bsize = jdata['stat_bsize']

    data = get_thermo(thermo_file)
    lx, lxe = block_avg(data[:, 8], skip = stat_skip, block_size = stat_bsize)
    ly, lye = block_avg(data[:, 9], skip = stat_skip, block_size = stat_bsize)
    lz, lze = block_avg(data[:,10], skip = stat_skip, block_size = stat_bsize)
    xy, xye = block_avg(data[:,11], skip = stat_skip, block_size = stat_bsize)
    xz, xze = block_avg(data[:,12], skip = stat_skip, block_size = stat_bsize)
    yz, yze = block_avg(data[:,13], skip = stat_skip, block_size = stat_bsize)
    # print('~~~', lx , ly, lz , xy, xz, yz)
    
    last_dump = get_last_dump(dump_file).split('\n')
    sys_data = system_data(last_dump)
    sys_data['cell'][0][0] = lx
    sys_data['cell'][1][1] = ly
    sys_data['cell'][2][2] = lz
    sys_data['cell'][1][0] = xy
    sys_data['cell'][2][0] = xz
    sys_data['cell'][2][1] = yz

    conf_lmp = from_system_data(sys_data)
    return conf_lmp

def extract(job_dir, output) :
    dump_file = os.path.join(job_dir, 'dump.avgposi')
    if os.path.isfile(dump_file) :
        print('# found dump.avgposi, use it')
    else :
        dump_file = os.path.join(job_dir, 'dump.equi')
        assert(os.path.isfile(dump_file))
        print('# found dump.equi, use it')        
    last_dump = get_last_dump(dump_file).split('\n')
    for idx in range(len(last_dump)) :
        ii = last_dump[idx]
        if 'ITEM: ATOMS' in ii :
            ii = ii.replace('f_ap[1]', 'x')
            ii = ii.replace('f_ap[2]', 'y')
            ii = ii.replace('f_ap[3]', 'z')
        last_dump[idx] = ii
    sys_data = system_data(last_dump)
    conf_lmp = from_system_data(sys_data)
    with open(output, 'w') as f:
        f.write(conf_lmp)
    # open(output, 'w').write(conf_lmp)

def make_task(iter_name, jdata, ens=None, temp=None, pres=None, if_dump_avg_posi=None, npt_dir=None):
    # jfile_path = os.path.abspath(jfile)
    # with open(jfile, 'r') as f:
    #     jdata = json.load(f)

    equi_args = [
        Argument("equi_conf", str),
        Argument("mass_map", list, alias=['model_mass_map']),
        Argument("model", str),
        Argument("nsteps", int),
        Argument("timestep", float, alias=['dt']),
        Argument("ens", str),
        Argument("temp", int),
        Argument("pres", int),
        Argument("tau_t", float),
        Argument("tau_p", float),
        Argument("thermo_freq", int, alias=['stat_freq']),
        Argument("dump_freq", int),
        Argument("stat_skip", int),
        Argument("stat_bsize", int),
        Argument("if_dump_avg_posi", bool, optional=True, default=False),
        Argument("is_water", bool, optional=True, default=False, alias=['if_water']),
        Argument("if_meam", bool, optional=True, default=False),
        Argument("meam_model", list, optional=True, default=False),
    ]

    equi_format = Argument("equi", dict, equi_args)

    equi_kwargs_settings = create_dict_not_empty_key(
        ens=ens,
        temp=temp,
        pres=pres,
        if_dump_avg_posi=if_dump_avg_posi,
        npt_dir=npt_dir
    )

    equi_pre_settings = jdata.copy()
    equi_pre_settings.update(equi_kwargs_settings)

    equi_settings = equi_format.normalize_value(equi_pre_settings)

    task_abs_dir = create_path(iter_name)

    if npt_dir is not None:
        npt_avg_conf_lmp = npt_equi_conf(npt_dir)
        with open(os.path.join(task_abs_dir, 'npt_avg.lmp'), 'w') as f:
            f.write(npt_avg_conf_lmp)
        equi_settings['equi_conf'] = 'npt_avg.lmp'
    else:
        equi_conf = equi_settings['equi_conf']
        relative_link_file(equi_conf, task_abs_dir)
        equi_settings['equi_conf'] = os.path.basename(equi_conf)

    model = equi_settings['model']
    if model:
        relative_link_file(model, task_abs_dir)
        equi_settings['model'] = os.path.basename(model)

    if_meam = equi_settings.get('if_meam', None)
    meam_model = equi_settings.get('meam_model', None)
    if if_meam:
        relative_link_file(meam_model['library'], task_abs_dir)
        relative_link_file(meam_model['potential'], task_abs_dir)


    with open(os.path.join(task_abs_dir, 'jdata.json'), 'w') as f:
        json.dump(jdata, f, indent=4)
    with open(os.path.join(task_abs_dir, 'equi_kwargs_setting.json'), 'w') as f:
        json.dump(equi_kwargs_settings, f, indent=4)
    with open(os.path.join(task_abs_dir, 'equi_settings.json'), 'w') as f:
        json.dump(equi_settings, f, indent=4)

    lmp_str = gen_equi_lammps_input(nsteps=equi_settings['nsteps'], 
        thermo_freq=equi_settings['thermo_freq'], 
        dump_freq=equi_settings['dump_freq'],
        mass_map=equi_settings['mass_map'],
        temp=equi_settings['temp'],
        tau_t=equi_settings['tau_t'],
        tau_p=equi_settings['tau_p'],
        equi_conf=equi_settings['equi_conf'],
        model=equi_settings['model'],
        timestep=equi_settings['timestep'],
        if_dump_avg_posi=equi_settings['if_dump_avg_posi'],
        ens=equi_settings['ens'],
        pres=equi_settings['pres'],
        if_meam=equi_settings['if_meam'], 
        meam_model=equi_settings['meam_model'])


    with open(os.path.join(task_abs_dir, 'in.lammps'), 'w') as fp :
        fp.write(lmp_str)
    return equi_settings


def water_bond(iter_name, skip = 1) :
    fdump = os.path.join(iter_name, 'dump.equi')
    with open(fdump, 'r') as f:
        lines = f.read().split('\n')
    # lines = open(fdump).read().split('\n')
    sections = []
    for ii in range(len(lines)) :
        if 'ITEM: TIMESTEP' in lines[ii] :
            sections.append(ii)
    sections.append(len(lines)-1) 
    all_rr = []
    all_tt = []
    for ii in range(skip, len(sections)-1) :
        sys_data = system_data(lines[sections[ii]:sections[ii+1]])
        atype = sys_data['atom_types']
        posis = sys_data['coordinates']
        cell  = sys_data['cell']
        bonds = compute_bonds(cell, atype, posis)
        if ii == skip : 
            bonds_0 = bonds 
        else :
            if bonds_0 != bonds :
                print('proton trans detected at frame %d' % ii)
        rr = []
        tt = []
        for ii in range(len(bonds)) :
            if atype[ii] == 1 :
                i_idx = ii
                j_idx = bonds[i_idx][0]
                k_idx = bonds[i_idx][1]
                drj = posi_diff(cell, posis[i_idx], posis[j_idx])
                drk = posi_diff(cell, posis[i_idx], posis[k_idx])
                ndrj = np.linalg.norm(drj)
                ndrk = np.linalg.norm(drk)
                rr.append(ndrj)
                rr.append(ndrk)
                tt.append(np.arccos(np.dot(drj,drk) / (ndrj*ndrk)))
        all_rr += rr
        all_tt += tt
    print('# statistics over %d frames %d angles' % (len(sections)-1, len(all_tt)))
    return (np.average(all_rr)), (np.average(all_tt))


def _compute_thermo (lmplog, natoms, stat_skip, stat_bsize) :
    # print(3939, natoms)
    data = get_thermo(lmplog)
    ea, ee = block_avg(data[:, 3], skip = stat_skip, block_size = stat_bsize)
    ha, he = block_avg(data[:, 4], skip = stat_skip, block_size = stat_bsize)
    ta, te = block_avg(data[:, 5], skip = stat_skip, block_size = stat_bsize)
    pa, pe = block_avg(data[:, 6], skip = stat_skip, block_size = stat_bsize)
    va, ve = block_avg(data[:, 7], skip = stat_skip, block_size = stat_bsize)
    lxx, lxxe = block_avg(data[:, 8], skip = stat_skip, block_size = stat_bsize)
    lyy, lyye = block_avg(data[:, 9], skip = stat_skip, block_size = stat_bsize)
    lzz, lzze = block_avg(data[:,10], skip = stat_skip, block_size = stat_bsize)
    lxy, lxye = block_avg(data[:,11], skip = stat_skip, block_size = stat_bsize)
    lxz, lxze = block_avg(data[:,12], skip = stat_skip, block_size = stat_bsize)
    lyz, lyze = block_avg(data[:,13], skip = stat_skip, block_size = stat_bsize)
    pxx, pxxe = block_avg(data[:,14], skip = stat_skip, block_size = stat_bsize)
    pyy, pyye = block_avg(data[:,15], skip = stat_skip, block_size = stat_bsize)
    pzz, pzze = block_avg(data[:,16], skip = stat_skip, block_size = stat_bsize)
    pxy, pxye = block_avg(data[:,17], skip = stat_skip, block_size = stat_bsize)
    pxz, pxze = block_avg(data[:,18], skip = stat_skip, block_size = stat_bsize)
    pyz, pyze = block_avg(data[:,19], skip = stat_skip, block_size = stat_bsize)
    thermo_info = {}
    thermo_info['p'] = pa
    thermo_info['p_err'] = pe
    thermo_info['v'] = va / natoms 
    thermo_info['v_err'] = ve / np.sqrt(natoms)
    thermo_info['e'] = ea / natoms 
    thermo_info['e_err'] = ee / np.sqrt(natoms)
    thermo_info['t'] = ta
    thermo_info['t_err'] = te
    thermo_info['h'] = ha / natoms 
    thermo_info['h_err'] = he / np.sqrt(natoms)
    unit_cvt = 1e5 * (1e-10**3) / pc.electron_volt
    thermo_info['pv'] = pa * va * unit_cvt / natoms
    thermo_info['pv_err'] = pe * va * unit_cvt / np.sqrt(natoms)
    thermo_info['lxx'] = lxx
    thermo_info['lxx_err'] = lxxe
    thermo_info['lyy'] = lyy
    thermo_info['lyy_err'] = lyye
    thermo_info['lzz'] = lzz
    thermo_info['lzz_err'] = lzze
    thermo_info['lxy'] = lxy
    thermo_info['lxy_err'] = lxye
    thermo_info['lxz'] = lxz
    thermo_info['lxz_err'] = lxze
    thermo_info['lyz'] = lyz
    thermo_info['lyz_err'] = lyze
    thermo_info['pxx'] = pxx
    thermo_info['pxx_err'] = pxxe
    thermo_info['pyy'] = pyy
    thermo_info['pyy_err'] = pyye
    thermo_info['pzz'] = pzz
    thermo_info['pzz_err'] = pzze
    thermo_info['pxy'] = pxy
    thermo_info['pxy_err'] = pxye
    thermo_info['pxz'] = pxz
    thermo_info['pxz_err'] = pxze
    thermo_info['pyz'] = pyz
    thermo_info['pyz_err'] = pyze
    return thermo_info

def _print_thermo_info(info, more_head = '') :
    ptr  = '# thermodynamics  %20s %20s  %s\n' % ('value', 'err', more_head)
    ptr += '# E        [eV]:  %20.8f %20.8f\n' % (info['e'], info['e_err'])
    ptr += '# H        [eV]:  %20.8f %20.8f\n' % (info['h'], info['h_err'])
    ptr += '# T         [K]:  %20.8f %20.8f\n' % (info['t'], info['t_err'])
    ptr += '# P       [bar]:  %20.8f %20.8f\n' % (info['p'], info['p_err'])
    ptr += '# V       [A^3]:  %20.8f %20.8f\n' % (info['v'], info['v_err'])
    ptr += '# PV       [eV]:  %20.8f %20.8f\n' % (info['pv'], info['pv_err'])
    ptr += '# Lxx       [A]:  %20.8f %20.8f\n' % (info['lxx'], info['lxx_err'])
    ptr += '# Lyy       [A]:  %20.8f %20.8f\n' % (info['lyy'], info['lyy_err'])
    ptr += '# Lzz       [A]:  %20.8f %20.8f\n' % (info['lzz'], info['lzz_err'])
    ptr += '# Lxy       [A]:  %20.8f %20.8f\n' % (info['lxy'], info['lxy_err'])
    ptr += '# Lxz       [A]:  %20.8f %20.8f\n' % (info['lxz'], info['lxz_err'])
    ptr += '# Lyz       [A]:  %20.8f %20.8f\n' % (info['lyz'], info['lyz_err'])
    ptr += '# Pxx     [bar]:  %20.8f %20.8f\n' % (info['pxx'], info['pxx_err'])
    ptr += '# Pyy     [bar]:  %20.8f %20.8f\n' % (info['pyy'], info['pyy_err'])
    ptr += '# Pzz     [bar]:  %20.8f %20.8f\n' % (info['pzz'], info['pzz_err'])
    ptr += '# Pxy     [bar]:  %20.8f %20.8f\n' % (info['pxy'], info['pxy_err'])
    ptr += '# Pxz     [bar]:  %20.8f %20.8f\n' % (info['pxz'], info['pxz_err'])
    ptr += '# Pyz     [bar]:  %20.8f %20.8f\n' % (info['pyz'], info['pyz_err'])
    rho = (18 * 1e-3 / (info['v'] * pc.Avogadro * pc.angstrom**3))
    rho_err = (info['v'] / (info['v'] - info['v_err'] ) - 1) * rho
    ptr += '# water density [kg/m^3] : %10.5f (%10.5f)' % (rho, rho_err)
    print(ptr)
    return ptr

def post_task(iter_name, natoms = None, is_water = None) :
    # j_file = os.path.join(iter_name, 'in.json')
    j_file = os.path.join(iter_name, 'equi_settings.json')
    with open(j_file, 'r') as f:
        jdata = json.load(f)
    if natoms == None :
        equi_conf = get_task_file_abspath(iter_name, jdata['equi_conf'])
        natoms = get_natoms(equi_conf)
        if 'copies' in jdata :
            natoms *= np.prod(jdata['copies'])
    if is_water is None:
        is_water=jdata.get('is_water', False)
    else:
        pass
    if is_water is True:
        nmols = natoms // 3
    elif is_water is False:
        nmols = natoms
    else:
        raise RuntimeError('must specify key "is_water" of  bool value  in jdata')
    stat_skip = jdata['stat_skip']
    stat_bsize = jdata['stat_bsize']
    log_file = os.path.join(iter_name, 'log.lammps')
    info = _compute_thermo(log_file, nmols, stat_skip, stat_bsize)
    ptr = _print_thermo_info(info)

    info_dict = info.copy()
    out_lmp = os.path.abspath(os.path.join(iter_name, 'out.lmp'))
    info_dict['out_lmp'] = out_lmp
    info_dict['job_dir'] = iter_name

    with open(os.path.join(iter_name, 'result'), 'w') as f:
        f.write(ptr)
    # open(').write(ptr).close()
    with open(os.path.join(iter_name, 'result.json'), 'w') as f:
        json.dump(info_dict, f, indent=4)
        # f.write(json.dump(info))
    # open(, 'w').write(json.dumps(info)).close()
    return info_dict

def _main ():
    parser = argparse.ArgumentParser(
        description="Equilibrium simulation")
    subparsers = parser.add_subparsers(title='Valid subcommands', dest='command')

    parser_gen = subparsers.add_parser('gen', help='Generate a job')
    parser_gen.add_argument('PARAM', type=str ,
                            help='json parameter file')
    parser_gen.add_argument('-e','--ensemble', type=str,
                            help='the ensemble of the simulation')
    parser_gen.add_argument('-t','--temperature', type=float,
                            help='the temperature of the system')
    parser_gen.add_argument('-p','--pressure', type=float,
                            help='the pressure of the system')
    parser_gen.add_argument('-a','--avg-posi', action = 'store_true',
                            help='dump the average position of atoms')
    parser_gen.add_argument('-c','--conf-npt', type=str,
                            help='use conf computed from NPT simulation')
    parser_gen.add_argument('-o','--output', type=str, default = 'new_job',
                            help='the output folder for the job')
    # parser_gen.add_argument("-z", "--meam", help="whether use meam instead of dp", action="store_true")

    parser_comp = subparsers.add_parser('extract', help= 'Extract the conf')
    parser_comp.add_argument('JOB', type=str ,
                             help='folder of the job')
    parser_comp.add_argument('-o','--output', type=str, default = 'conf.lmp',
                             help='output conf file name')

    parser_stat = subparsers.add_parser('stat-bond', help= 'Statistic of the bonds')
    parser_stat.add_argument('JOB', type=str ,
                             help='folder of the job')
    parser_stat.add_argument('-s','--skip', type=int, default = 1,
                             help='skip this number of frames')

    parser_stat = subparsers.add_parser('compute', help= 'Compute thermodynamics')
    parser_stat.add_argument('JOB', type=str ,
                             help='folder of the job')

    args = parser.parse_args()

    
    if args.command is None :
        parser.print_help()
        exit
    if args.command == 'gen' :
        jdata = json.load(open(args.PARAM, 'r'))
        # jfile = os.path.abspath(args.PARAM)
        make_task(args.output, jdata, args.ensemble, args.temperature, args.pressure, args.avg_posi, args.conf_npt,)
    elif args.command == 'extract' :
        extract(args.JOB, args.output)
    elif args.command == 'stat-bond' :
        b, a = water_bond(args.JOB, args.skip)
        print(b, a/np.pi*180)
    elif args.command == 'compute' :
        post_task(args.JOB)


if __name__ == '__main__' :
    _main()



import os, json, shutil
import numpy as np
import unittest
from context import dpti
from unittest.mock import MagicMock, patch, PropertyMock
from dpti.lib.utils import get_file_md5

class TestHtiMakeTask(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        os.mkdir('tmp_hti/')

    def setUp(self):
        self.maxDiff = None
        self.test_dir = 'tmp_hti'
        self.benchmark_dir = 'benchmark_hti'

    @patch('numpy.random')
    def test_deepmd_three_step(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        test_name = 'three_step'
        benchmark_dir = os.path.join(self.benchmark_dir, test_name)
        test_dir = os.path.join(self.test_dir, test_name)

        json_file = os.path.join(benchmark_dir, 'jdata.json')
        with open(json_file) as f:
            jdata = json.load(f)
        # print(dpti.ti)
        # print(dir(dpti.ti))
        dpti.hti.make_tasks(iter_name=test_dir, jdata=jdata, switch='three-step')
        check_file_list = [ 
            'conf.lmp', 
            'graph.pb',
            '00.lj_on/task.000004/conf.lmp',
            '00.lj_on/task.000004/in.lammps',
            '01.deep_on/task.000005/graph.pb',
            '01.deep_on/task.000005/in.lammps',
            '02.spring_off/task.000006/lambda.out',
            '02.spring_off/task.000006/in.lammps'
        ]
        for file in check_file_list:
            f1 = os.path.join(benchmark_dir, file)
            f2 = os.path.join(test_dir, file)
            self.assertEqual(get_file_md5(f1), get_file_md5(f2), msg=(f1,f2))


    @patch('numpy.random')
    def test_deepmd_two_step(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        test_name = 'two_step'
        benchmark_dir = os.path.join(self.benchmark_dir, test_name)
        test_dir = os.path.join(self.test_dir, test_name)

        json_file = os.path.join(benchmark_dir, 'jdata.json')
        with open(json_file) as f:
            jdata = json.load(f)
        # print(dpti.ti)
        # print(dir(dpti.ti))
        dpti.hti.make_tasks(iter_name=test_dir, jdata=jdata, switch='two-step')
        check_file_list = [ 
            'conf.lmp', 
            'graph.pb',
            '00.deep_on/task.000005/conf.lmp',
            '00.deep_on/task.000005/in.lammps',
            '01.spring_off/task.000006/graph.pb',
            '01.spring_off/task.000006/in.lammps',
        ]
        for file in check_file_list:
            f1 = os.path.join(benchmark_dir, file)
            f2 = os.path.join(test_dir, file)
            self.assertEqual(get_file_md5(f1), get_file_md5(f2), msg=(f1,f2))

    @patch('numpy.random')
    def test_deepmd_one_step(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        test_name = 'one_step'
        benchmark_dir = os.path.join(self.benchmark_dir, test_name)
        test_dir = os.path.join(self.test_dir, test_name)

        json_file = os.path.join(benchmark_dir, 'jdata.json')
        with open(json_file) as f:
            jdata = json.load(f)
        dpti.hti.make_tasks(iter_name=test_dir, jdata=jdata, switch='one-step')
        check_file_list = [ 
            'conf.lmp', 
            'graph.pb',
            'task.000006/graph.pb',
            'task.000006/in.lammps',
        ]
        for file in check_file_list:
            f1 = os.path.join(benchmark_dir, file)
            f2 = os.path.join(test_dir, file)
            self.assertEqual(get_file_md5(f1), get_file_md5(f2), msg=(f1,f2))

    @patch('numpy.random')
    def test_meam_three_step(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        test_name = 'three_step_meam'
        benchmark_dir = os.path.join(self.benchmark_dir, test_name)
        test_dir = os.path.join(self.test_dir, test_name)

        json_file = os.path.join(benchmark_dir, 'jdata.json')
        with open(json_file) as f:
            jdata = json.load(f)
        # print(dpti.ti)
        # print(dir(dpti.ti))
        dpti.hti.make_tasks(iter_name=test_dir, jdata=jdata, switch='three-step')
        check_file_list = [ 
            'conf.lmp', 
            '00.lj_on/task.000004/conf.lmp',
            '00.lj_on/task.000004/in.lammps',
            '01.deep_on/task.000005/library_18Metals.meam',
            '01.deep_on/task.000005/in.lammps',
            '02.spring_off/task.000006/Sn_18Metals.meam',
            '02.spring_off/task.000006/in.lammps'
        ]
        for file in check_file_list:
            f1 = os.path.join(benchmark_dir, file)
            f2 = os.path.join(test_dir, file)
            self.assertEqual(get_file_md5(f1), get_file_md5(f2), msg=(f1,f2))


    @classmethod
    def tearDownClass(cls):
        shutil.rmtree('tmp_hti/')

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import numpy as np

def _get_block (lines, keys) :
    for idx in range(len(lines)) :
        if keys in lines[idx] :
            break
    if idx == len(lines) - 1:
        return None
    idx_s = idx+2
    idx = idx_s
    ret = []
    while True :        
        if len(lines[idx].split()) == 0 :
            break
        else :
            ret.append(lines[idx])
        idx += 1
    return ret

def lmpbox2box(lohi, tilt) :
    xy = tilt[0]
    xz = tilt[1]
    yz = tilt[2]
    orig = np.array([lohi[0][0], lohi[1][0], lohi[2][0]])
    lens = []
    for dd in range(3) :
        lens.append(lohi[dd][1] - lohi[dd][0])
    xx = [lens[0], 0, 0]
    yy = [xy, lens[1], 0]
    zz=  [xz, yz, lens[2]]
    return orig, np.array([xx, yy, zz])

def box2lmpbox(orig, box) :
    lohi = np.zeros([3,2])
    for dd in range(3) :
        lohi[dd][0] = orig[dd]
    tilt = np.zeros(3)
    tilt[0] = box[1][0]
    tilt[1] = box[2][0]
    tilt[2] = box[2][1]
    lens = np.zeros(3) 
    lens[0] = box[0][0]
    lens[1] = box[1][1]
    lens[2] = box[2][2]
    for dd in range(3) :
        lohi[dd][1] = lohi[dd][0] + lens[dd]
    return lohi, tilt

def get_atoms(lines) :
    return _get_block(lines, 'Atoms')

def get_natoms(lines) :
    for ii in lines :
        if 'atoms' in ii :
            return int(ii.split()[0])
    return None

def get_natomtypes(lines) :
    for ii in lines :
        if 'atom types' in ii :
            return int(ii.split()[0])
    return None

def _atom_info_mol(line) :
    vec = line.split()
    # idx, mole_type, atom_type, charge, x, y, z
    return int(vec[0]), int(vec[1]), int(vec[2]), float(vec[3]), float(vec[4]), float(vec[5]), float(vec[6])

def _atom_info_atom(line) :
    vec = line.split()
    # idx, atom_type, x, y, z
    return int(vec[0]), int(vec[1]), float(vec[2]), float(vec[3]), float(vec[4])

def get_natoms_vec(lines) :
    atype = get_atype(lines)
    natoms_vec = []
    natomtypes = get_natomtypes(lines)
    for ii in range(natomtypes) :
        natoms_vec.append(sum(atype == ii+1))
    assert (sum(natoms_vec) == get_natoms(lines))
    return natoms_vec

def get_atype(lines) :
    alines = get_atoms(lines) 
    atype = []
    for ii in alines :
        # idx, mt, at, q, x, y, z = _atom_info_mol(ii)
        idx, at, x, y, z = _atom_info_atom(ii)
        atype.append(at)
    return np.array(atype, dtype = int)

def get_posi(lines) :
    atom_lines = get_atoms(lines)
    posis = []
    for ii in atom_lines :
        # posis.append([float(jj) for jj in ii.split()[4:7]])
        posis.append([float(jj) for jj in ii.split()[2:5]])
    return np.array(posis)

def get_lmpbox(lines) :
    box_info = []
    tilt = np.zeros(3)
    for ii in lines :
        if 'xlo' in ii and 'xhi' in ii :
            box_info.append([float(ii.split()[0]), float(ii.split()[1])])
            break
    for ii in lines :
        if 'ylo' in ii and 'yhi' in ii :
            box_info.append([float(ii.split()[0]), float(ii.split()[1])])
            break
    for ii in lines :
        if 'zlo' in ii and 'zhi' in ii :
            box_info.append([float(ii.split()[0]), float(ii.split()[1])])
            break
    for ii in lines :
        if 'xy' in ii and 'xz' in ii and 'yz' in ii :
            tilt = np.array([float(jj) for jj in ii.split()[0:3]])
    return box_info, tilt


def system_data(lines) :
    system = {}
    system['atom_numbs'] = get_natoms_vec(lines)
    system['atom_names'] = []
    for ii in range(len(system['atom_numbs'])) :
        system['atom_names'].append('Type_%d' % ii)
    lohi, tilt = get_lmpbox(lines)
    orig, cell = lmpbox2box(lohi, tilt)
    system['orig'] = np.array(orig)
    system['cell'] = np.array(cell)
    natoms = sum(system['atom_numbs'])
    system['atom_types'] = get_atype(lines)
    system['coordinates'] = get_posi(lines)
    return system

def to_system_data(lines) :
    return system_data(lines)

def from_system_data(system) :
    ret = ''
    ret += '\n'
    natoms = sum(system['atom_numbs'])
    ntypes = len(system['atom_numbs'])    
    ret += '%d atoms\n' % natoms
    ret += '%d atom types\n' % ntypes
    ret += '0 %.8f xlo xhi\n' % system['cell'][0][0]
    ret += '0 %.8f ylo yhi\n' % system['cell'][1][1]
    ret += '0 %.8f zlo zhi\n' % system['cell'][2][2]
    ret += '%.8f %.8f %.8f xy xz yz\n' % \
    (system['cell'][1][0], system['cell'][2][0], system['cell'][2][1])
    ret += '\n'
    ret += 'Atoms # atomic\n'
    ret += '\n'
    for ii in range(natoms) :
        ret += '%d %d %.8f %.8f %.8f\n' % \
               (ii+1,
                system['atom_types'][ii],
                system['coordinates'][ii][0] - system['orig'][0],
                system['coordinates'][ii][1] - system['orig'][1],
                system['coordinates'][ii][2] - system['orig'][2]
        )
    return ret


if __name__ == '__main__' :
    fname = 'water-SPCE.data'
    lines = open(fname).read().split('\n')
    bonds, tilt = get_lmpbox(lines)
    # print(bonds, tilt)
    orig, box = lmpbox2box(bonds, tilt)
    # print(orig, box)
    bonds1, tilt1 = box2lmpbox(orig, box)
    # print(bonds1, tilt1)
    print(bonds1 - bonds)
    print(tilt1 - tilt)
    print(box)
    print(get_atype(lines))
    print(get_posi(lines))



import os, textwrap
import numpy as np
import unittest
from context import dpti
from potential_common import soft_param, soft_param_three_element, meam_model

class TestFfSpring(unittest.TestCase):
    def setUp(self) :
        self.maxDiff = None

    def test_lj_on(self):
        input = dict(lamb=0.075,
            model="graph.pb",
            m_spring_k=[118.71],
            step="lj_on",
            sparam=soft_param,
            if_meam=False,
            meam_model=None
        )
        ret1 = textwrap.dedent("""\
        # --------------------- FORCE FIELDS ---------------------
        variable        EPSILON equal 0.030000
        pair_style      lj/cut/soft 1.000000 0.500000 6.000000
        pair_coeff      1 1 ${EPSILON} 2.493672 0.500000
        fix             tot_pot all adapt/fep 0 pair lj/cut/soft epsilon * * v_LAMBDA scale yes
        compute         e_diff all fep ${TEMP} pair lj/cut/soft epsilon * * v_EPSILON
        group           type_1 type 1
        fix             l_spring_1 type_1 spring/self 1.1871000000e+02
        fix_modify      l_spring_1 energy yes
        variable        l_spring equal f_l_spring_1
        """)
        ret2 = dpti.hti._ff_soft_lj(**input)
        # print('--------')
        # # print(ret1)
        # print('--------')
        # # print(ret2)
        # print('--------')
        self.assertEqual(ret1, ret2)

    def test_deep_on(self):
        input = dict(lamb=0.075,
            model="graph.pb",
            m_spring_k=[118.71],
            step="deep_on",
            sparam=soft_param,
            if_meam=False,
            meam_model=None
        )
        ret1 = textwrap.dedent("""\
        # --------------------- FORCE FIELDS ---------------------
        variable        EPSILON equal 0.030000
        variable        ONE equal 1
        pair_style      hybrid/overlay deepmd graph.pb lj/cut/soft 1.000000 0.500000 6.000000
        pair_coeff      * * deepmd
        pair_coeff      1 1 lj/cut/soft ${EPSILON} 2.493672 0.500000
        fix             tot_pot all adapt/fep 0 pair deepmd scale * * v_LAMBDA
        compute         e_diff all fep ${TEMP} pair deepmd scale * * v_ONE
        group           type_1 type 1
        fix             l_spring_1 type_1 spring/self 1.1871000000e+02
        fix_modify      l_spring_1 energy yes
        variable        l_spring equal f_l_spring_1
        """)
        ret2 = dpti.hti._ff_soft_lj(**input)
        # print('--------')
        # print(ret1)
        # print('--------')
        # print(ret2)
        # print('--------')
        self.assertEqual(ret1, ret2)

    def test_spring_off(self):
        input = dict(lamb=0.075,
            model="graph.pb",
            m_spring_k=[118.71],
            step="spring_off",
            sparam=soft_param,
            if_meam=False,
            meam_model=None
        )
        ret1 = textwrap.dedent("""\
        # --------------------- FORCE FIELDS ---------------------
        variable        EPSILON equal 0.030000
        variable        INV_EPSILON equal -${EPSILON}
        pair_style      hybrid/overlay deepmd graph.pb lj/cut/soft 1.000000 0.500000 6.000000
        pair_coeff      * * deepmd
        pair_coeff      1 1 lj/cut/soft ${EPSILON} 2.493672 0.500000
        fix             tot_pot all adapt/fep 0 pair lj/cut/soft epsilon * * v_INV_LAMBDA scale yes
        compute         e_diff all fep ${TEMP} pair lj/cut/soft epsilon * * v_INV_EPSILON
        group           type_1 type 1
        fix             l_spring_1 type_1 spring/self 1.0980675000e+02
        fix_modify      l_spring_1 energy yes
        variable        l_spring equal f_l_spring_1
        """)
        ret2 = dpti.hti._ff_soft_lj(**input)
        self.assertEqual(ret1, ret2)

# def test_me
    def test_meam_deep_on(self):
        input = dict(lamb=0.075,
            model="graph.pb",
            m_spring_k=[118.71],
            step="deep_on",
            sparam=soft_param,
            if_meam=True,
            meam_model=meam_model
        )
        ret1 = textwrap.dedent("""\
        # --------------------- FORCE FIELDS ---------------------
        variable        EPSILON equal 0.030000
        variable        ONE equal 1
        pair_style      hybrid/overlay meam lj/cut/soft 1.000000 0.500000 6.000000
        pair_coeff      * * meam library_18Metals.meam Sn Sn_18Metals.meam Sn
        pair_coeff      1 1 lj/cut/soft ${EPSILON} 2.493672 0.500000
        fix             tot_pot all adapt/fep 0 pair meam scale * * v_LAMBDA
        compute         e_diff all fep ${TEMP} pair meam scale * * v_ONE
        group           type_1 type 1
        fix             l_spring_1 type_1 spring/self 1.1871000000e+02
        fix_modify      l_spring_1 energy yes
        variable        l_spring equal f_l_spring_1
        """)
        ret2 = dpti.hti._ff_soft_lj(**input)
        # print('--------')
        # print(ret1)
        # print('--------')
        # print(ret2)
        # print('--------')
        self.assertEqual(ret1, ret2)

    def test_meam_spring_off(self):
        input = dict(lamb=0.075,
            model="graph.pb",
            m_spring_k=[118.71],
            step="spring_off",
            sparam=soft_param,
            if_meam=True,
            meam_model=meam_model
        )
        
        ret1 = textwrap.dedent("""\
        # --------------------- FORCE FIELDS ---------------------
        variable        EPSILON equal 0.030000
        variable        INV_EPSILON equal -${EPSILON}
        pair_style      hybrid/overlay meam lj/cut/soft 1.000000 0.500000 6.000000
        pair_coeff      * * meam library_18Metals.meam Sn Sn_18Metals.meam Sn
        pair_coeff      1 1 lj/cut/soft ${EPSILON} 2.493672 0.500000
        fix             tot_pot all adapt/fep 0 pair lj/cut/soft epsilon * * v_INV_LAMBDA scale yes
        compute         e_diff all fep ${TEMP} pair lj/cut/soft epsilon * * v_INV_EPSILON
        group           type_1 type 1
        fix             l_spring_1 type_1 spring/self 1.0980675000e+02
        fix_modify      l_spring_1 energy yes
        variable        l_spring equal f_l_spring_1
        """)
        ret2 = dpti.hti._ff_soft_lj(**input)
        # print('--------')
        # print(ret1)
        # print('--------')
        # print(ret2)
        # print('--------')
        self.assertEqual(ret1, ret2)


#     def test_spring_var_spring(self):
#         input = dict(lamb=0.075, m_spring_k=[118.71],
#             var_spring=True)
#         ret1 = textwrap.dedent("""\
#         group           type_1 type 1
#         fix             l_spring_1 type_1 spring/self 1.0980675000e+02
#         fix_modify      l_spring_1 energy yes
#         variable        l_spring equal f_l_spring_1
#         """)
#         ret2 = dpti.hti._ff_spring(**input)
#         self.assertEqual(ret1, ret2)

#     def test_spring_multiple_element(self):
#         input = dict(lamb=0.075, m_spring_k=[118.71, 207.2],
#             var_spring=False)
#         ret1 = textwrap.dedent("""\
#         group           type_1 type 1
#         group           type_2 type 2
#         fix             l_spring_1 type_1 spring/self 1.1871000000e+02
#         fix_modify      l_spring_1 energy yes
#         fix             l_spring_2 type_2 spring/self 2.0720000000e+02
#         fix_modify      l_spring_2 energy yes
#         variable        l_spring equal f_l_spring_1+f_l_spring_2
#         """)
#         ret2 = dpti.hti._ff_spring(**input)
#         self.assertEqual(ret1, ret2)

# def test_spring_var_spring_multiple_element(self):
#         input = dict(lamb=0.20, m_spring_k=[118.71, 207.2],
#             var_spring=False)
#         ret1 = textwrap.dedent("""\
#         group           type_1 type 1
#         group           type_2 type 2
#         fix             l_spring_1 type_1 spring/self 9.4968000000e+01
#         fix_modify      l_spring_1 energy yes
#         fix             l_spring_2 type_2 spring/self 1.6576000000e+02
#         fix_modify      l_spring_2 energy yes
#         variable        l_spring equal f_l_spring_1+f_l_spring_2
#         """)
#         ret2 = dpti.hti._ff_spring(**input)
#         self.assertEqual(ret1, ret2)


if __name__ == '__main__':
    unittest.main()



import json, os
from airflow import DAG
from airflow.operators.dummy import DummyOperator
from airflow.operators.python_operator import PythonOperator
from airflow.operators.subdag_operator import SubDagOperator
from airflow.operators.python import get_current_context
from airflow.decorators import task
from airflow.utils.dates import days_ago
from airflow.utils.task_group import TaskGroup
from datetime import datetime, timedelta
from airflow.exceptions import AirflowSkipException, DagNotFound, DagRunAlreadyExists, DagRunNotFound, AirflowException,  AirflowFailException
# from airflow.api.client.local_client import Client
from airflow.utils.state import State

from airflow.api.client.local_client import Client
from airflow.models import Variable, DagRun
from numpy.core.fromnumeric import var
# from dpdispatcher.
from dpdispatcher.submission import Submission, Task

import time
import uuid
from dpti import gdi

from dpti.gdi import gdi_main_loop

# default_args = {'owner': 'airflow',
#                 'start_date': datetime(2018, 1, 1)
#                 }

# BASE_DAG_NAME='dag_dpti_gdi_v8'
# MAX_LOOP_NUM = 30

class GDIDAGFactory:
    default_args = {'owner': 'airflow',
            'start_date': datetime(2018, 1, 1)
    }

    dagargs = {
        'default_args': default_args,
        'schedule_interval': None,
    }
    def __init__(self, gdi_name, dag_work_base):
        self.gdi_name = gdi_name
        self.dag_loop_name = self.gdi_name + '_gdi_loop_dag'
        self.dag_main_name = self.gdi_name + '_gdi_main_dag'
        self.var_name = self.gdi_name + '_dv_dh'
        self.dag_work_base = dag_work_base
        self.main_dag = self.create_main_dag()
        self.loop_dag = self.create_loop_dag()

    def create_main_dag(self):
        dag_name = self.dag_loop_name
        var_name = self.var_name
        work_base = self.dag_work_base
        @task()
        def dpti_gdi_main_prepare(**kwargs):
            # context = get_current_context()
            # dag_run = context['params']
            # work_base = dag_run['work_base']
            # Variable.set(self.var_name, 'run')
            prepare_return = True
            return prepare_return
        
        # @task()
        @task(retries=2, retry_delay=timedelta(minutes=1))
        def dpti_gdi_main_loop(prepare_return, **kwargs):
            # context = get_current_context()
            # dag_run = context['params']
            # work_base = dag_run['work_base']
            # work_base = work_base
            print('debug:prepare_return', prepare_return)
            with open(os.path.join(work_base, 'machine.json'), 'r') as f:
                mdata = json.load(f)

            with open(os.path.join(work_base, 'pb.json'), 'r') as f:
                jdata = json.load(f)

            with open(os.path.join(work_base, 'gdidata.json'), 'r') as f:
                gdidata_dict = json.load(f)

            output_dir = os.path.join(work_base, 'new_job/')
            gdidata_dict['output'] = output_dir

            # workflow =
            gdi_workflow = GDIWorkflow(var_name=var_name, 
                dag_name=dag_name)

            gdi_main_loop(jdata=jdata, 
                mdata=mdata,
                gdidata_dict=gdidata_dict,
                gdidata_cli={}, 
                workflow=gdi_workflow
            )
            # return True          
            # Variable.set(self.var_name, 'run')
            loop_return = True
            return loop_return

        @task()
        def dpti_gdi_main_end(loop_return, **kwargs):            
            # Variable.set(self.var_name, 'run')
            end_return = True
            return end_return

        main_dag = DAG(self.dag_main_name, **self.__class__.dagargs)
        with main_dag:
            prepare_return = dpti_gdi_main_prepare()
            loop_return = dpti_gdi_main_loop(prepare_return)
            end_return = dpti_gdi_main_end(loop_return)

        return main_dag

    def create_loop_dag(self):
        @task(multiple_outputs=True)
        def dpti_gdi_loop_prepare():
            # Variable.set(self.var_name, 'run')

            context = get_current_context()
            dag_run = context['params']
            task0_dict = dag_run['task_dict_list'][0]
            task1_dict = dag_run['task_dict_list'][1]

            submission_dict = dag_run['submission_dict']
            # prepare_return = True
            # return (task0_dict, task1_dict)
            return {'task0_dict':task0_dict, 'task1_dict':task1_dict}
            # return (task0_dict, task1_dict)

        # @task()
        @task(retries=2, retry_delay=timedelta(minutes=1))
        def dpti_gdi_loop_md(task_dict):
            context = get_current_context()
            dag_run = context['params']

            submission_dict = dag_run['submission_dict']
            print('submission_dict', submission_dict)
            mdata = dag_run['mdata']
            print('mdata', mdata)
            print('debug:task_dict', task_dict)

            machine = Machine.load_from_dict(mdata['machine'])
            resources = Resources.load_from_dict(mdata['resources'])
            submission = Submission.deserialize(
                submission_dict=submission_dict,
                machine=machine
            )
            submission.resources = resources

            submission.register_task(task=Task.deserialize(task_dict=task_dict))
            submission.run_submission()
            # md_return = prepare_return
            return True

        @task()
        def dpti_gdi_loop_end(task0_return, task1_return):
            end_return = True
            # Variable.set(self.var_name, 'end')
            return end_return

        loop_dag = DAG(self.dag_loop_name, **self.__class__.dagargs)
        with loop_dag:
            tasks_dict = dpti_gdi_loop_prepare()
            task0_return = dpti_gdi_loop_md(tasks_dict['task0_dict'])
            task1_return = dpti_gdi_loop_md(tasks_dict['task1_dict'])
            end_return = dpti_gdi_loop_end(task0_return, task1_return)
        return loop_dag


class GDIWorkflow:
    def __init__(self, dag_name, var_name):
        self.dag_name = dag_name
        self.var_name = var_name
        self.run_id = None

    def get_dag_run_state(self):
        if self.run_id is None:
            raise DagRunNotFound(f"dag_id {self.dag_name}; {self.run_id}")

        dag_runs = DagRun.find(
            dag_id=self.dag_name,
            run_id=self.run_id
        )
        
        return dag_runs[0].state if dag_runs else None

    def wait_until_end(self):
        while True:
            dag_run_state = self.get_dag_run_state()
            if dag_run_state == State.SUCCESS:
                print(f"dag_run_state: {dag_run_state}")
                break
            elif dag_run_state == State.RUNNING:
                print(f"dag_run_state: {dag_run_state}")
                time.sleep(30)
            else:
                raise AirflowFailException(f"subdag dag_run fail dag_id:{self.dag_name}; run_id:{self.run_id};")
        return dag_run_state

    def trigger_loop(self, submission, task_list, mdata):
        # loop_num = None
        c = Client(None, None)
        submission_dict = submission.serialize()
        task_dict_list = [task.serialize() for task in task_list]
        submission_hash = submission.submission_hash
        self.run_id = f"dag_run_{submission_hash}"
        try:
            c.trigger_dag(dag_id=self.dag_name, run_id=self.run_id,
                conf={'submission_dict': submission_dict, 'task_dict_list':task_dict_list, 'mdata':mdata}
            ) 
        except DagRunAlreadyExists:
            dag_run_state = self.get_dag_run_state()
            if dag_run_state == State.FAILED:
                raise AirflowFailException(f"subdag dag_run fail dag_id:{self.dag_name}; run_id:{self.run_id};")
            else:
                print(f"continue from old dag_run {self.run_id}")
        loop_return = self.wait_until_end()
        return loop_return








import os, textwrap
import numpy as np
import unittest
from context import dpti

class TestEquiThermoSetting(unittest.TestCase):
    def setUp(self) :
        self.maxDiff = None

    def test_gen_equi_thermo_settings(self):
        input = dict(timestep=0.002)
        ret1 = textwrap.dedent("""\
        # --------------------- MD SETTINGS ----------------------
        neighbor        1.0 bin
        timestep        0.002000
        thermo          ${THERMO_FREQ}
        compute         allmsd all msd
        thermo_style    custom step ke pe etotal enthalpy temp press vol lx ly lz xy xz yz pxx pyy pzz pxy pxz pyz c_allmsd[*]
        """)
        ret2 = dpti.equi.gen_equi_thermo_settings(**input)
        self.assertEqual(ret1, ret2)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
import os, sys, json
import numpy
from unittest.mock import MagicMock, patch, PropertyMock
from context import dpti
# assert os.path.isdir('equi_benchmark') is False

with patch.object(numpy.random, 'randint', return_value=7858) as mock_method:
# @patch('numpy.random')
# def gen_equi_benchmark(patch_random):
    # patch_random.randint = MagicMock(return_value=7858)
    with open('npt.json', 'r') as f:
        equi_npt_benchmark_settings = json.load(f)
    dpti.equi.make_task(iter_name='equi_benchmark/npt', jdata=equi_npt_benchmark_settings)

    with open('npt_meam.json', 'r') as f:
        equi_npt_meam_benchmark_settings = json.load(f)
    dpti.equi.make_task(iter_name='equi_benchmark/npt_meam', jdata=equi_npt_meam_benchmark_settings)

# gen_equi_benchmark()



#!/usr/bin/env python3

import os, sys
import numpy as np
lib_path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(lib_path)
import lmp

def _get_block (lines, key) :
    for idx in range(len(lines)) :
        if ('ITEM: ' + key) in lines[idx] :
            break
    idx_s = idx + 1
    for idx in range(idx_s, len(lines)) :
        if ('ITEM: ') in lines[idx] :
            break
    idx_e = idx
    if idx_e == len(lines)-1 :
        idx_e += 1
    return lines[idx_s:idx_e], lines[idx_s-1]

def get_atype(lines) :
    blk, head = _get_block(lines, 'ATOMS')
    keys = head.split()
    id_idx = keys.index('id') - 2
    tidx = keys.index('type') - 2
    atype = []
    for ii in blk :
        atype.append([int(ii.split()[id_idx]), int(ii.split()[tidx])])
    atype.sort()
    atype = np.array(atype, dtype = int)
    return atype[:,1]

def get_natoms(lines) :
    blk, head = _get_block(lines, 'NUMBER OF ATOMS')
    return int(blk[0])

def get_natomtypes(lines) :
    atype = get_atype(lines)
    return max(atype)

def get_natoms_vec(lines) :
    atype = get_atype(lines)
    natoms_vec = []
    natomtypes = get_natomtypes(lines)
    for ii in range(natomtypes) :
        natoms_vec.append(sum(atype == ii+1))
    assert (sum(natoms_vec) == get_natoms(lines))
    return natoms_vec

def get_posi(lines) :
    blk, head = _get_block(lines, 'ATOMS')
    keys = head.split()
    id_idx = keys.index('id') - 2
    xidx = keys.index('x') - 2
    yidx = keys.index('y') - 2
    zidx = keys.index('z') - 2
    sel = (xidx, yidx, zidx)
    posis = []
    for ii in blk :
        words = ii.split()
        posis.append([float(words[id_idx]), float(words[xidx]), float(words[yidx]), float(words[zidx])])
    posis.sort()
    posis = np.array(posis)     
    return posis[:,1:4]

def get_dumpbox(lines) :
    blk, h = _get_block(lines, 'BOX BOUNDS')
    bounds = np.zeros([3,2])
    tilt = np.zeros([3])
    for dd in range(3) :
        info = [float(jj) for jj in blk[dd].split()]
        bounds[dd][0] = info[0]
        bounds[dd][1] = info[1]
        tilt[dd] = info[2]
    return bounds, tilt

def dumpbox2box(bounds, tilt) :
    xy = tilt[0]
    xz = tilt[1]
    yz = tilt[2]
    xlo = bounds[0][0] - min(0.0,xy,xz,xy+xz)
    xhi = bounds[0][1] - max(0.0,xy,xz,xy+xz)
    ylo = bounds[1][0] - min(0.0,yz)
    yhi = bounds[1][1] - max(0.0,yz)
    zlo = bounds[2][0]
    zhi = bounds[2][1]
    info = [[xlo, xhi], [ylo, yhi], [zlo, zhi]]
    return lmp.lmpbox2box(info, tilt)

def box2dumpbox(orig, box) :
    lohi, tilt = lmp.box2lmpbox(orig, box)
    xy = tilt[0]
    xz = tilt[1]
    yz = tilt[2]
    bounds = np.zeros([3,2])
    bounds[0][0] = lohi[0][0] + min(0.0,xy,xz,xy+xz)
    bounds[0][1] = lohi[0][1] + max(0.0,xy,xz,xy+xz)
    bounds[1][0] = lohi[1][0] + min(0.0,yz)
    bounds[1][1] = lohi[1][1] + max(0.0,yz)
    bounds[2][0] = lohi[2][0]
    bounds[2][1] = lohi[2][1]
    return bounds, tilt

def system_data(lines) :
    system = {}
    system['atom_numbs'] = get_natoms_vec(lines)
    system['atom_names'] = []
    for ii in range(len(system['atom_numbs'])) :
        system['atom_names'].append('TYPE_%d' % ii)
    bounds, tilt = get_dumpbox(lines)
    orig, cell = dumpbox2box(bounds, tilt)
    system['orig'] = np.array(orig)
    system['cell'] = np.array(cell)
    natoms = sum(system['atom_numbs'])
    system['atom_types'] = get_atype(lines)
    system['coordinates'] = get_posi(lines)
    return system


def split_traj(dump_lines) :
    marks = []
    for idx,ii in enumerate(dump_lines) :
        if 'ITEM: TIMESTEP' in ii :
            marks.append(idx) 
    if len(marks) == 0 :
        return None
    elif len(marks) == 1 :
        return [dump_lines]
    else :
        block_size = marks[1] - marks[0]
        ret = []
        for ii in marks :
            ret.append(dump_lines[ii:ii+block_size])
        # for ii in range(len(marks)-1): 
        #     assert(marks[ii+1] - marks[ii] == block_size)
        return ret    
    return None


if __name__ == '__main__' :
    fname = 'dump.hti'
    lines = open(fname).read().split('\n')
    # print(get_natoms(lines))
    # print(get_natomtypes(lines))
    # print(get_natoms_vec(lines))
    posi = get_posi(lines)
    dbox, tilt = get_dumpbox(lines)
    orig, box = dumpbox2box(dbox, tilt)
    dbox1, tilt1 = box2dumpbox(orig, box)
    print(dbox - dbox1)
    print(tilt - tilt1)
    print(orig)
    print(box)
    np.savetxt('tmp.out', posi - orig, fmt='%.6f')
    print(system_data(lines))



import os, textwrap
import numpy as np
import unittest
from context import dpti
# from potential_common import soft_param, soft_param_three_element, meam_model
# from dpti.lib.lammps import get_natoms, get_thermo, get_last_dump
# from dpti.lib.dump import from_system_data
from potential_common import soft_param, meam_model
from dpti.hti import _gen_lammps_input
from numpy.testing import assert_almost_equal
from unittest.mock import MagicMock, patch, PropertyMock

class TestHtiGenLammpsInput(unittest.TestCase):
    def setUp(self) :
        self.maxDiff = None

    @patch('numpy.random')
    def test_deepmd_lj_on(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        input = dict(conf_file='conf.lmp', 
                        mass_map=[118.71],
                        lamb=0.075,
                        model="graph.pb",
                        m_spring_k=[2.3742,],
                        nsteps=500000,
                        timestep=0.002,
                        ens='npt',
                        temp=400.0,
                        pres = 1.0, 
                        tau_t = 0.1,
                        tau_p = 0.5,
                        thermo_freq = 100, 
                        copies = None,
                        crystal = 'vega', 
                        sparam = soft_param,
                        switch = 'three-step',
                        step = 'lj_on',
                        if_meam = False,
                        meam_model = None)
        ret1 = textwrap.dedent("""\
        clear
        # --------------------- VARIABLES-------------------------
        variable        NSTEPS          equal 500000
        variable        THERMO_FREQ     equal 100
        variable        DUMP_FREQ       equal 100
        variable        TEMP            equal 400.000000
        variable        PRES            equal 1.000000
        variable        TAU_T           equal 0.100000
        variable        TAU_P           equal 0.500000
        variable        LAMBDA          equal 7.5000000000e-02
        variable        INV_LAMBDA      equal 9.2500000000e-01
        # ---------------------- INITIALIZAITION ------------------
        units           metal
        boundary        p p p
        atom_style      atomic
        # --------------------- ATOM DEFINITION ------------------
        box             tilt large
        read_data       conf.lmp
        change_box      all triclinic
        mass            1 118.710000
        # --------------------- FORCE FIELDS ---------------------
        variable        EPSILON equal 0.030000
        pair_style      lj/cut/soft 1.000000 0.500000 6.000000
        pair_coeff      1 1 ${EPSILON} 2.493672 0.500000
        fix             tot_pot all adapt/fep 0 pair lj/cut/soft epsilon * * v_LAMBDA scale yes
        compute         e_diff all fep ${TEMP} pair lj/cut/soft epsilon * * v_EPSILON
        group           type_1 type 1
        fix             l_spring_1 type_1 spring/self 2.3742000000e+00
        fix_modify      l_spring_1 energy yes
        variable        l_spring equal f_l_spring_1
        # --------------------- MD SETTINGS ----------------------
        neighbor        1.0 bin
        timestep        0.002
        thermo          ${THERMO_FREQ}
        compute         allmsd all msd
        thermo_style    custom step ke pe etotal enthalpy temp press vol v_l_spring c_e_diff[1] c_allmsd[*]
        thermo_modify   format 9 %.16e
        thermo_modify   format 10 %.16e
        # dump            1 all custom ${DUMP_FREQ} dump.hti id type x y z vx vy vz
        fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}
        # --------------------- INITIALIZE -----------------------
        velocity        all create ${TEMP} 7858
        group           first id 1
        fix             fc first recenter INIT INIT INIT
        fix             fm first momentum 1 linear 1 1 1
        velocity        first zero linear
        # --------------------- RUN ------------------------------
        run             ${NSTEPS}
        write_data      out.lmp
        """)
        ret2 = _gen_lammps_input(**input)
        # print(ret2)
        self.assertEqual(ret1, ret2)


    @patch('numpy.random')
    def test_deepmd_deep_on(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        input = dict ( conf_file='conf.lmp', 
                        mass_map=[118.71],
                        lamb=0.075,
                        model="graph.pb",
                        m_spring_k=[2.3742,],
                        nsteps=500000,
                        timestep=0.002,
                        ens='npt',
                        temp=400.0,
                        pres = 1.0, 
                        tau_t = 0.1,
                        tau_p = 0.5,
                        thermo_freq = 100, 
                        copies = None,
                        crystal = 'vega', 
                        sparam = soft_param,
                        switch = 'three-step',
                        step = 'deep_on',
                        if_meam = False,
                        meam_model = None)
        ret1 = textwrap.dedent("""\
        clear
        # --------------------- VARIABLES-------------------------
        variable        NSTEPS          equal 500000
        variable        THERMO_FREQ     equal 100
        variable        DUMP_FREQ       equal 100
        variable        TEMP            equal 400.000000
        variable        PRES            equal 1.000000
        variable        TAU_T           equal 0.100000
        variable        TAU_P           equal 0.500000
        variable        LAMBDA          equal 7.5000000000e-02
        variable        INV_LAMBDA      equal 9.2500000000e-01
        # ---------------------- INITIALIZAITION ------------------
        units           metal
        boundary        p p p
        atom_style      atomic
        # --------------------- ATOM DEFINITION ------------------
        box             tilt large
        read_data       conf.lmp
        change_box      all triclinic
        mass            1 118.710000
        # --------------------- FORCE FIELDS ---------------------
        variable        EPSILON equal 0.030000
        variable        ONE equal 1
        pair_style      hybrid/overlay deepmd graph.pb lj/cut/soft 1.000000 0.500000 6.000000
        pair_coeff      * * deepmd
        pair_coeff      1 1 lj/cut/soft ${EPSILON} 2.493672 0.500000
        fix             tot_pot all adapt/fep 0 pair deepmd scale * * v_LAMBDA
        compute         e_diff all fep ${TEMP} pair deepmd scale * * v_ONE
        group           type_1 type 1
        fix             l_spring_1 type_1 spring/self 2.3742000000e+00
        fix_modify      l_spring_1 energy yes
        variable        l_spring equal f_l_spring_1
        # --------------------- MD SETTINGS ----------------------
        neighbor        1.0 bin
        timestep        0.002
        thermo          ${THERMO_FREQ}
        compute         allmsd all msd
        thermo_style    custom step ke pe etotal enthalpy temp press vol v_l_spring c_e_diff[1] c_allmsd[*]
        thermo_modify   format 9 %.16e
        thermo_modify   format 10 %.16e
        # dump            1 all custom ${DUMP_FREQ} dump.hti id type x y z vx vy vz
        fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}
        # --------------------- INITIALIZE -----------------------
        velocity        all create ${TEMP} 7858
        group           first id 1
        fix             fc first recenter INIT INIT INIT
        fix             fm first momentum 1 linear 1 1 1
        velocity        first zero linear
        # --------------------- RUN ------------------------------
        run             ${NSTEPS}
        write_data      out.lmp
        """)
        ret2 = _gen_lammps_input(**input)
        self.assertEqual(ret1, ret2)

    @patch('numpy.random')
    def test_deepmd_lj_off(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        input = dict ( conf_file='conf.lmp', 
                        mass_map=[118.71],
                        lamb=0.075,
                        model="graph.pb",
                        m_spring_k=[2.3742,],
                        nsteps=500000,
                        timestep=0.002,
                        ens='npt',
                        temp=400.0,
                        pres = 1.0, 
                        tau_t = 0.1,
                        tau_p = 0.5,
                        thermo_freq = 100, 
                        copies = None,
                        crystal = 'vega', 
                        sparam = soft_param,
                        switch = 'three-step',
                        step = 'spring_off',
                        if_meam = False,
                        meam_model = None)
        ret1 = textwrap.dedent("""\
        clear
        # --------------------- VARIABLES-------------------------
        variable        NSTEPS          equal 500000
        variable        THERMO_FREQ     equal 100
        variable        DUMP_FREQ       equal 100
        variable        TEMP            equal 400.000000
        variable        PRES            equal 1.000000
        variable        TAU_T           equal 0.100000
        variable        TAU_P           equal 0.500000
        variable        LAMBDA          equal 7.5000000000e-02
        variable        INV_LAMBDA      equal 9.2500000000e-01
        # ---------------------- INITIALIZAITION ------------------
        units           metal
        boundary        p p p
        atom_style      atomic
        # --------------------- ATOM DEFINITION ------------------
        box             tilt large
        read_data       conf.lmp
        change_box      all triclinic
        mass            1 118.710000
        # --------------------- FORCE FIELDS ---------------------
        variable        EPSILON equal 0.030000
        variable        INV_EPSILON equal -${EPSILON}
        pair_style      hybrid/overlay deepmd graph.pb lj/cut/soft 1.000000 0.500000 6.000000
        pair_coeff      * * deepmd
        pair_coeff      1 1 lj/cut/soft ${EPSILON} 2.493672 0.500000
        fix             tot_pot all adapt/fep 0 pair lj/cut/soft epsilon * * v_INV_LAMBDA scale yes
        compute         e_diff all fep ${TEMP} pair lj/cut/soft epsilon * * v_INV_EPSILON
        group           type_1 type 1
        fix             l_spring_1 type_1 spring/self 2.1961350000e+00
        fix_modify      l_spring_1 energy yes
        variable        l_spring equal f_l_spring_1
        # --------------------- MD SETTINGS ----------------------
        neighbor        1.0 bin
        timestep        0.002
        thermo          ${THERMO_FREQ}
        compute         allmsd all msd
        thermo_style    custom step ke pe etotal enthalpy temp press vol v_l_spring c_e_diff[1] c_allmsd[*]
        thermo_modify   format 9 %.16e
        thermo_modify   format 10 %.16e
        # dump            1 all custom ${DUMP_FREQ} dump.hti id type x y z vx vy vz
        fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}
        # --------------------- INITIALIZE -----------------------
        velocity        all create ${TEMP} 7858
        group           first id 1
        fix             fc first recenter INIT INIT INIT
        fix             fm first momentum 1 linear 1 1 1
        velocity        first zero linear
        # --------------------- RUN ------------------------------
        run             ${NSTEPS}
        write_data      out.lmp
        """)
        ret2 = _gen_lammps_input(**input)
        self.assertEqual(ret1, ret2)

    @patch('numpy.random')
    def test_meam_deep_on(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        input = dict ( conf_file='conf.lmp', 
                        mass_map=[118.71],
                        lamb=0.075,
                        model=None,
                        m_spring_k=[2.3742,],
                        nsteps=500000,
                        timestep=0.002,
                        ens='npt',
                        temp=400.0,
                        pres = 1.0, 
                        tau_t = 0.1,
                        tau_p = 0.5,
                        thermo_freq = 100, 
                        copies = None,
                        crystal = 'vega', 
                        sparam = soft_param,
                        switch = 'three-step',
                        step = 'deep_on',
                        if_meam = True,
                        meam_model = meam_model)
        ret1 = textwrap.dedent("""\
        clear
        # --------------------- VARIABLES-------------------------
        variable        NSTEPS          equal 500000
        variable        THERMO_FREQ     equal 100
        variable        DUMP_FREQ       equal 100
        variable        TEMP            equal 400.000000
        variable        PRES            equal 1.000000
        variable        TAU_T           equal 0.100000
        variable        TAU_P           equal 0.500000
        variable        LAMBDA          equal 7.5000000000e-02
        variable        INV_LAMBDA      equal 9.2500000000e-01
        # ---------------------- INITIALIZAITION ------------------
        units           metal
        boundary        p p p
        atom_style      atomic
        # --------------------- ATOM DEFINITION ------------------
        box             tilt large
        read_data       conf.lmp
        change_box      all triclinic
        mass            1 118.710000
        # --------------------- FORCE FIELDS ---------------------
        variable        EPSILON equal 0.030000
        variable        ONE equal 1
        pair_style      hybrid/overlay meam lj/cut/soft 1.000000 0.500000 6.000000
        pair_coeff      * * meam library_18Metals.meam Sn Sn_18Metals.meam Sn
        pair_coeff      1 1 lj/cut/soft ${EPSILON} 2.493672 0.500000
        fix             tot_pot all adapt/fep 0 pair meam scale * * v_LAMBDA
        compute         e_diff all fep ${TEMP} pair meam scale * * v_ONE
        group           type_1 type 1
        fix             l_spring_1 type_1 spring/self 2.3742000000e+00
        fix_modify      l_spring_1 energy yes
        variable        l_spring equal f_l_spring_1
        # --------------------- MD SETTINGS ----------------------
        neighbor        1.0 bin
        timestep        0.002
        thermo          ${THERMO_FREQ}
        compute         allmsd all msd
        thermo_style    custom step ke pe etotal enthalpy temp press vol v_l_spring c_e_diff[1] c_allmsd[*]
        thermo_modify   format 9 %.16e
        thermo_modify   format 10 %.16e
        # dump            1 all custom ${DUMP_FREQ} dump.hti id type x y z vx vy vz
        fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}
        # --------------------- INITIALIZE -----------------------
        velocity        all create ${TEMP} 7858
        group           first id 1
        fix             fc first recenter INIT INIT INIT
        fix             fm first momentum 1 linear 1 1 1
        velocity        first zero linear
        # --------------------- RUN ------------------------------
        run             ${NSTEPS}
        write_data      out.lmp
        """)
        ret2 = _gen_lammps_input(**input)
        # print(ret2)
        self.assertEqual(ret1, ret2)

    @patch('numpy.random')
    def test_meam_spring_off(self, patch_random):
        patch_random.randint = MagicMock(return_value=7858)
        input = dict ( conf_file='conf.lmp', 
                        mass_map=[118.71],
                        lamb=0.075,
                        model=None,
                        m_spring_k=[2.3742,],
                        nsteps=500000,
                        timestep=0.002,
                        ens='npt',
                        temp=400.0,
                        pres = 1.0, 
                        tau_t = 0.1,
                        tau_p = 0.5,
                        thermo_freq = 100, 
                        copies = None,
                        crystal = 'vega', 
                        sparam = soft_param,
                        switch = 'three-step',
                        step = 'spring_off',
                        if_meam = True,
                        meam_model = meam_model)
        ret1 = textwrap.dedent("""\
        clear
        # --------------------- VARIABLES-------------------------
        variable        NSTEPS          equal 500000
        variable        THERMO_FREQ     equal 100
        variable        DUMP_FREQ       equal 100
        variable        TEMP            equal 400.000000
        variable        PRES            equal 1.000000
        variable        TAU_T           equal 0.100000
        variable        TAU_P           equal 0.500000
        variable        LAMBDA          equal 7.5000000000e-02
        variable        INV_LAMBDA      equal 9.2500000000e-01
        # ---------------------- INITIALIZAITION ------------------
        units           metal
        boundary        p p p
        atom_style      atomic
        # --------------------- ATOM DEFINITION ------------------
        box             tilt large
        read_data       conf.lmp
        change_box      all triclinic
        mass            1 118.710000
        # --------------------- FORCE FIELDS ---------------------
        variable        EPSILON equal 0.030000
        variable        INV_EPSILON equal -${EPSILON}
        pair_style      hybrid/overlay meam lj/cut/soft 1.000000 0.500000 6.000000
        pair_coeff      * * meam library_18Metals.meam Sn Sn_18Metals.meam Sn
        pair_coeff      1 1 lj/cut/soft ${EPSILON} 2.493672 0.500000
        fix             tot_pot all adapt/fep 0 pair lj/cut/soft epsilon * * v_INV_LAMBDA scale yes
        compute         e_diff all fep ${TEMP} pair lj/cut/soft epsilon * * v_INV_EPSILON
        group           type_1 type 1
        fix             l_spring_1 type_1 spring/self 2.1961350000e+00
        fix_modify      l_spring_1 energy yes
        variable        l_spring equal f_l_spring_1
        # --------------------- MD SETTINGS ----------------------
        neighbor        1.0 bin
        timestep        0.002
        thermo          ${THERMO_FREQ}
        compute         allmsd all msd
        thermo_style    custom step ke pe etotal enthalpy temp press vol v_l_spring c_e_diff[1] c_allmsd[*]
        thermo_modify   format 9 %.16e
        thermo_modify   format 10 %.16e
        # dump            1 all custom ${DUMP_FREQ} dump.hti id type x y z vx vy vz
        fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}
        # --------------------- INITIALIZE -----------------------
        velocity        all create ${TEMP} 7858
        group           first id 1
        fix             fc first recenter INIT INIT INIT
        fix             fm first momentum 1 linear 1 1 1
        velocity        first zero linear
        # --------------------- RUN ------------------------------
        run             ${NSTEPS}
        write_data      out.lmp
        """)
        ret2 = _gen_lammps_input(**input)
        # print(ret2)
        self.assertEqual(ret1, ret2)

    def test_raise_err(self):
        input = dict ( conf_file='conf.lmp', 
                        mass_map=[118.71],
                        lamb=0.075,
                        model="graph.pb",
                        m_spring_k=[2.3742,],
                        nsteps=500000,
                        timestep=0.002,
                        ens='npt',
                        temp=400.0,
                        pres = 1.0, 
                        tau_t = 0.1,
                        tau_p = 0.5,
                        thermo_freq = 100, 
                        copies = None,
                        crystal = 'vega', 
                        sparam = soft_param,
                        switch = 'three-step',
                        step = 'spring_off',
                        if_meam = False,
                        meam_model = None)
        
        input2 = input.copy()
        input2['step'] = 'foo'
        with self.assertRaises(RuntimeError):
            _gen_lammps_input(**input2)

        
        input3 = input.copy()
        input3['switch'] = 'bar'
        with self.assertRaises(RuntimeError):
            _gen_lammps_input(**input3)

        input4 = input.copy()
        input4['ens'] = 'baz'
        with self.assertRaises(RuntimeError):
            _gen_lammps_input(**input4)

